"""
The ``context`` module provides :py:class:`opendp.context.Context` and supporting utilities.

For more context, see :ref:`context in the User Guide <context-user-guide>`.

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp
"""

import logging
from typing import Any, Callable, Optional, Sequence, Type, Union, MutableMapping
import importlib
from inspect import signature
from functools import partial
from opendp.combinators import (
    make_fix_delta,
    make_approximate,
    make_fully_adaptive_composition,
    make_privacy_filter,
    make_pureDP_to_zCDP,
    make_adaptive_composition,
    make_zCDP_to_approxDP,
)
from opendp.domains import atom_domain, vector_domain, with_margin
from opendp.extras._utilities import supports_partial, to_then
from opendp.measurements import make_laplace, make_gaussian
from opendp.measures import (
    approximate,
    max_divergence,
    zero_concentrated_divergence,
)
from opendp.metrics import (
    absolute_distance,
    change_one_distance,
    change_one_id_distance,
    discrete_distance,
    hamming_distance,
    insert_delete_distance,
    l1_distance,
    l2_distance,
    frame_distance,
    symmetric_distance,
    symmetric_id_distance,
)
from opendp.mod import (
    ApproximateDivergence,
    Domain,
    Measurement,
    Metric,
    _PartialConstructor,
    Odometer,
    OdometerQueryable,
    Queryable,
    Transformation,
    Measure,
    binary_search,
    binary_search_param,
)
from opendp.typing import RuntimeType
from opendp._lib import indent, import_optional_dependency
from opendp.extras.polars import Bound, LazyFrameQuery, Margin
from dataclasses import replace

__all__ = [
    "space_of",
    "domain_of",
    "metric_of",
    "loss_of",
    "unit_of",
    "Context",
    "Query",
    "Chain",
    "PartialChain",
    "register",
]


logger = logging.getLogger(__name__)


# a dictionary of "constructor name" -> (constructor_function, is_partial)
# "constructor name" is the name of the constructor without the "make_" prefix
# constructor_function is the partial version if is_partial is True
constructors = {}
for module_name in ["transformations", "measurements"]:
    module = importlib.import_module(f"opendp.{module_name}")
    for name in module.__all__:
        if not name.startswith("make_"):
            continue
        partial_name = "then_" + name[5:]
        make_func = getattr(module, name)

        is_partial = partial_name in module.__all__
        constructor = getattr(module, partial_name if is_partial else name)

        constructors[name[5:]] = constructor, is_partial


def register(
    constructor: Callable[..., Union[Transformation, Measurement]],
    name: Optional[str] = None,
):
    """Register a constructor function to be used in the Context API.

    If the constructor supports partial application (first two arguments are ``input_domain`` and ``input_metric``),
    then the input domain and input metric are omitted when called via the Context API.

    Constructor requirements:

    * The constructor must return a transformation or measurement.
    * If name is None, the constructor's name must start with ``make_``.

    :param constructor: The constructor function to register.
    :param name: The name to register the constructor under in the Context API. If None, the name will be derived from the constructor's name.
    """
    if name is None:
        if not constructor.__name__.startswith("make_"):
            raise ValueError(
                f"constructor.__name__ must start with 'make_', found {constructor.__name__}"
            )
        name = constructor.__name__[5:]

    if name in constructors:
        raise ValueError(
            f"'{name}' is already registered in the Context API. Please choose a different name."
        )

    if supports_partial(constructor):
        constructors[name] = to_then(constructor), True
    else:
        constructors[name] = constructor, False


def space_of(T, M=None, infer: bool = False) -> tuple[Domain, Metric]:
    """A shorthand for building a metric space.

    A metric space consists of a domain and a metric.

    >>> import opendp.prelude as dp
    ...
    >>> dp.space_of(list[int])
    (VectorDomain(AtomDomain(T=i32)), SymmetricDistance())
    >>> # the verbose form allows greater control:
    >>> (dp.vector_domain(dp.atom_domain(T=dp.i32)), dp.symmetric_distance())
    (VectorDomain(AtomDomain(T=i32)), SymmetricDistance())

    :param T: carrier type (the type of members in the domain)
    :param M: metric type
    :param infer: if True, ``T`` is an example of the sensitive dataset. Passing sensitive data may result in a privacy violation.
    """
    import opendp.typing as ty

    domain = domain_of(T, infer=infer)
    D = domain.type

    # choose a metric type if not set
    if M is None:
        if D.origin == "VectorDomain":  # type: ignore[union-attr]
            M = ty.SymmetricDistance
        elif D.origin == "AtomDomain" and ty.get_atom(D) in ty._NUMERIC_TYPES:  # type: ignore[union-attr]
            M = ty.AbsoluteDistance
        else:
            raise TypeError(
                f"no default metric for domain {D}. Please set `M`"
            )  # pragma: no cover

    # choose a distance type if not set
    if isinstance(M, ty.RuntimeType) and not M.args:
        M = M[ty.get_atom(D)]  # type: ignore[index]

    return domain, metric_of(M)


def domain_of(T, infer: bool = False) -> Domain:
    """Constructs an instance of a domain from carrier type ``T``, or from an example.

    Accepts a limited set of Python type expressions:

    >>> import opendp.prelude as dp
    >>> dp.domain_of(list[int])
    VectorDomain(AtomDomain(T=i32))

    As well as strings representing types in the underlying Rust syntax:

    >>> dp.domain_of('Vec<int>')
    VectorDomain(AtomDomain(T=i32))

    Dictionaries, optional types, and a range of primitive types are supported:

    >>> dp.domain_of(dict[str, int])
    MapDomain { key_domain: AtomDomain(T=String), value_domain: AtomDomain(T=i32) }

    .. TODO: Support python syntax for Option: https://github.com/opendp/opendp/issues/1389

    >>> dp.domain_of('Option<int>')  # Python's `Optional` is not supported.
    OptionDomain(AtomDomain(T=i32))
    >>> dp.domain_of(dp.i32)
    AtomDomain(T=i32)

    More complex types are not supported:

    >>> dp.domain_of(list[list[int]]) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    opendp.mod.OpenDPException:
      FFI("Inner domain of VectorDomain must be AtomDomain or ExtrinsicDomain (created via user_domain)")

    Alternatively, an example of the data can be provided, but note that passing sensitive data may result in a privacy violation:

    >>> dp.domain_of([1, 2, 3], infer=True)
    VectorDomain(AtomDomain(T=i32))

    :param T: carrier type
    :param infer: if True, ``T`` is an example of the sensitive dataset. Passing sensitive data may result in a privacy violation.
    """
    import opendp.typing as ty
    from opendp.domains import vector_domain, atom_domain, option_domain, map_domain

    if infer:
        pl = import_optional_dependency("polars", raise_error=False)
        if pl is not None and isinstance(T, pl.LazyFrame):
            from opendp.extras.polars import _lazyframe_domain_from_schema

            return _lazyframe_domain_from_schema(T.collect_schema())

    # normalize to a type descriptor
    if infer:
        T = ty.RuntimeType.infer(T)
    else:
        T = ty.RuntimeType.parse(T)

    # construct the domain
    if isinstance(T, ty.RuntimeType):
        if T.origin == "Vec":
            return vector_domain(domain_of(T.args[0]))
        if T.origin == "HashMap":
            return map_domain(domain_of(T.args[0]), domain_of(T.args[1]))
        if T.origin == "Option":
            return option_domain(domain_of(T.args[0]))

    if T in ty._PRIMITIVE_TYPES:
        return atom_domain(T=T)

    raise TypeError(f"unrecognized carrier type: {T}")  # pragma: no cover


def metric_of(M) -> Metric:
    """Constructs an instance of a metric from metric type ``M``.

    :param M: Metric type
    """
    import opendp.typing as ty
    import opendp.metrics as metrics

    if isinstance(M, Metric):
        return M
    M = ty.RuntimeType.parse(M)

    if isinstance(M, ty.RuntimeType):
        if M.origin == "AbsoluteDistance":
            return metrics.absolute_distance(T=M.args[0])
        if M.origin == "L1Distance":
            return metrics.l1_distance(T=M.args[0])
        if M.origin == "L2Distance":
            return metrics.l2_distance(T=M.args[0])

    if M == ty.HammingDistance:
        return metrics.hamming_distance()
    if M == ty.SymmetricDistance:
        return metrics.symmetric_distance()
    if M == ty.InsertDeleteDistance:
        return metrics.insert_delete_distance()
    if M == ty.ChangeOneDistance:
        return metrics.change_one_distance()
    if M == ty.DiscreteDistance:
        return metrics.discrete_distance()

    raise TypeError(f"unrecognized metric: {M}")  # pragma: no cover


def loss_of(
    epsilon: Optional[float] = None,
    delta: Optional[float] = None,
    rho: Optional[float] = None,
) -> tuple[Measure, Union[float, tuple[float, float]]]:
    """Constructs a privacy loss, consisting of a privacy measure and a privacy loss parameter.

    >>> import opendp.prelude as dp
    >>> dp.loss_of(epsilon=1.0)
    (MaxDivergence, 1.0)
    >>> dp.loss_of(epsilon=1.0, delta=1e-9)
    (Approximate(MaxDivergence), (1.0, 1e-09))
    >>> dp.loss_of(rho=1.0)
    (ZeroConcentratedDivergence, 1.0)

    :param epsilon: Parameter for pure ε-DP.
    :param delta: Parameter for δ-approximate DP.
    :param rho: Parameter for zero-concentrated ρ-DP.

    """

    def _range_warning(name, value, info_level, warn_level):
        if value > warn_level:
            if info_level == warn_level:
                logger.warning(f"{name} should be less than or equal to {warn_level}")
            else:
                logger.warning(
                    f"{name} should be less than or equal to {warn_level}, and is typically less than or equal to {info_level}"
                )
        elif value > info_level:
            logger.info(f"{name} is typically less than or equal to {info_level}")

    if [rho is None, epsilon is None].count(True) != 1:
        raise ValueError(
            "Either epsilon or rho must be specified, and they are mutually exclusive."
        )  # pragma: no cover

    if epsilon is not None:
        _range_warning("epsilon", epsilon, 1, 5)
        measure, loss = max_divergence(), epsilon

    if rho is not None:
        _range_warning("rho", rho, 0.25, 0.5)
        measure, loss = zero_concentrated_divergence(), rho

    if delta is None:
        return measure, loss
    else:
        _range_warning("delta", delta, 1e-6, 1e-6)
        return approximate(measure), (loss, delta)


def unit_of(
    *,
    contributions: Optional[Union[int, Sequence[Bound]]] = None,
    changes: Optional[int] = None,
    absolute: Optional[float] = None,
    l1: Optional[float] = None,
    l2: Optional[float] = None,
    local: Optional[bool] = None,
    identifier: Optional[str] = None,
    ordered: bool = False,
    U=None,
) -> tuple[Metric, Union[float, Sequence[Bound]]]:
    """Constructs a unit of privacy, consisting of a metric and a dataset distance.
    The parameters are mutually exclusive.

    >>> import opendp.prelude as dp
    >>> dp.unit_of(contributions=3)
    (SymmetricDistance(), 3)
    >>> dp.unit_of(l1=2.0)
    (L1Distance(f64), 2.0)

    :param contributions: Greatest number of records or identifiers a privacy unit may contribute to microdata
    :param changes: Greatest number of records a privacy unit may change in microdata
    :param absolute: Greatest absolute distance a privacy unit can influence a scalar aggregate data set
    :param l1: Greatest l1 distance a privacy unit can influence a vector aggregate data set
    :param l2: Greatest l2 distance a privacy unit can influence a vector aggregate data set
    :param local: Set to true if the dataset consists of one individual (for local-DP)
    :param identifier: Can be a column name or Polars expression.
    :param ordered: Set to ``True`` to use ``InsertDeleteDistance`` instead of ``SymmetricDistance``, or ``HammingDistance`` instead of ``ChangeOneDistance``.
    :param U: The type of the dataset distance."""

    def _is_distance(p, v):
        return p not in ["ordered", "U", "_is_distance", "identifier"] and v is not None

    if sum(1 for p, v in locals().items() if _is_distance(p, v)) != 1:
        raise ValueError("Must specify exactly one distance.")

    if local:
        if identifier is not None or ordered or U is not None:
            raise ValueError('"local" must be the only parameter')
        return discrete_distance(), int(local)

    if ordered:
        if contributions is None and changes is None:
            raise ValueError(
                '"ordered" is only valid with "changes" or "contributions"'
            )
        if identifier is not None:
            raise ValueError('"ordered" must be False when "identifier" is set')

    if contributions is not None:
        if identifier is None:
            metric = insert_delete_distance() if ordered else symmetric_distance()
        else:
            metric = symmetric_id_distance(identifier)

        if isinstance(contributions, Sequence):
            metric = frame_distance(metric)

        return metric, contributions

    if changes is not None:
        if identifier is None:
            metric = hamming_distance() if ordered else change_one_distance()
        else:
            metric = change_one_id_distance(identifier)

        return metric, changes

    if absolute is not None:
        metric = absolute_distance(T=RuntimeType.parse_or_infer(U, absolute))
        return metric, absolute
    if l1 is not None:
        metric = l1_distance(T=RuntimeType.parse_or_infer(U, l1))
        return metric, l1
    if l2 is not None:
        metric = l2_distance(T=RuntimeType.parse_or_infer(U, l2))
        return metric, l2

    raise Exception("No matching metric found")  # pragma: no cover


class Context(object):
    """A Context coordinates queries to an instance of a privacy :py:attr:`accountant`.

    It is recommended to use :py:meth:`Context.compositor` constructor instead of this one.

    :param accountant: The measurement used to spawn the queryable.
    :param queryable: Executes the queries and tracks the privacy expenditure.
    :param d_in: An upper bound on the distance between adjacent datasets.
    :param d_mids: A sequence of privacy losses for each query to be sent to the queryable. Used for compositors.
    :param d_out: An upper bound on the overall privacy loss. Used for filters.
    :param query_space: The input space of queries to the context. If not specified, it defaults to the input space of the compositor.
    """

    accountant: Union[Measurement, Odometer]
    """The accountant is the measurement used to spawn the queryable.
    It contains information about the queryable, 
    such as the input domain, input metric, and output measure expected of measurement queries sent to the queryable."""
    queryable: Union[Queryable, OdometerQueryable]
    """The queryable executes the queries and tracks the privacy expenditure."""

    def __init__(
        self,
        accountant: Measurement,
        queryable: Queryable,
        d_in: Union[float, Sequence[Bound]],
        d_mids: Optional[Union[list[float], list[tuple[float, float]]]] = None,
        d_out: Optional[Union[float, tuple[float, float]]] = None,
        query_space: Optional[tuple[Domain, Metric]] = None,
    ):
        self.accountant = accountant
        self.queryable = queryable
        self.d_in = d_in
        self.d_mids = d_mids
        self.d_out = d_out

        self.d_mids_consumed: Union[list[float], list[tuple[float, float]]] = []
        self.query_space = query_space

    def __repr__(self) -> str:
        return f"""Context(
    accountant = {indent(repr(self.accountant))},
    d_in       = {self.d_in},
    d_mids     = {self.d_mids},
    d_out      = {self.d_out})"""

    def deserialize_polars_plan(self, serialized_plan: bytes) -> "LazyFrameQuery":
        """
        Given a serialized Polars plan, wraps it with a LazyFrameQuery.
        See the :ref:`serialization documentation <lazyframe-serialization>`
        for context and full example.

        :param serialized_plan: A plan like that returned by ``query.polars_plan.serialize()``
        """
        import io

        pl = import_optional_dependency("polars")

        new_plan = pl.LazyFrame.deserialize(io.BytesIO(serialized_plan))
        new_query = self.query()
        if not isinstance(new_query, LazyFrameQuery):
            raise ValueError("'data' of context must be a LazyFrame")

        new_query.polars_plan = new_plan
        return new_query

    @staticmethod
    def compositor(
        data: Any,
        privacy_unit: tuple[Metric, Union[float, Sequence[Bound]]],
        privacy_loss: tuple[Measure, Any],
        split_evenly_over: Optional[int] = None,
        split_by_weights: Optional[Sequence[float]] = None,
        domain: Optional[Domain] = None,
        margins: Optional[Sequence[Margin]] = None,
    ) -> "Context":
        """Constructs a new context containing a sequential compositor with the given weights.

        If the domain is not specified, it will be inferred from the data.
        This makes the assumption that the structure of the data is public information.

        ``split_evenly_over`` and ``split_by_weights`` are mutually exclusive.

        When ``data`` is a Polars LazyFrame, queries are specified as a Polars compute plan.
        In addition, ``margins`` may be specified, which contain descriptors for the data under grouping.

        :param data: The data to be analyzed.
        :param privacy_unit: The privacy unit of the compositor.
        :param privacy_loss: The privacy loss of the compositor.
        :param split_evenly_over: The number of parts to evenly distribute the privacy loss.
        :param split_by_weights: A list of weights for each intermediate privacy loss.
        :param domain: The domain of the data.
        :param margins: Descriptors for grouped data."""
        if domain is None:
            domain = domain_of(data, infer=True)

        if margins:
            # allows dictionaries of {[by]: [margin]}
            if isinstance(margins, MutableMapping):
                from warnings import warn

                warn(
                    'Margin dicts should be replaced with lists, with the key supplied as the "by" kwarg',
                    DeprecationWarning,
                )
                margins = [replace(margin, by=by) for by, margin in margins.items()]

            for margin in margins:
                domain = with_margin(domain, margin)

        accountant, d_mids, d_out = _normalize_compositor(
            domain, privacy_unit, privacy_loss, split_evenly_over, split_by_weights
        )

        try:
            queryable = accountant(data)
        except TypeError as e:
            inferred_domain = domain_of(data, infer=True)
            if vector_domain(domain) == inferred_domain:
                # With Python 3.11, add_note is available, but pytest.raises doesn't see notes.
                e.args = (
                    e.args[0] + "; To fix, wrap domain kwarg with dp.vector_domain()",
                )
            raise e

        return Context(
            accountant=accountant,
            queryable=queryable,
            d_in=privacy_unit[1],
            d_mids=d_mids,
            d_out=d_out,
        )

    def __call__(self, query: Measurement):
        """Executes the given query on the context."""
        answer = self.queryable(query)

        # pop won't fail if queryable returned an answer
        self.d_mids_consumed.append(
            query.map(self.d_in) if self.d_mids is None else self.d_mids.pop(0)  # type: ignore[arg-type]
        )
        return answer

    def query(self, **kwargs) -> Union["Query", LazyFrameQuery]:
        """Starts a new Query to be executed in this context.

        If the context has been constructed with a sequence of privacy losses,
        the next loss will be used. Otherwise, the loss will be computed from the kwargs.

        :param kwargs: The privacy loss to use for the query. Passed directly into :py:func:`loss_of`.
        """
        d_query = None
        if self.d_mids is not None:
            if kwargs:
                raise ValueError(
                    f"Expected no privacy arguments but got {kwargs}"
                )  # pragma: no cover
            if not self.d_mids:
                raise ValueError(
                    "Privacy allowance has been exhausted"
                )  # pragma: no cover
            d_query = self.d_mids[0]
        elif kwargs:
            observed_measure, d_query = loss_of(**kwargs)

            expected_measure = self.accountant.output_measure
            if observed_measure != expected_measure:
                msg = f"Expected output measure {expected_measure} but got {observed_measure}."
                if (
                    isinstance(expected_measure, ApproximateDivergence)
                    and "delta" not in kwargs
                ):
                    msg += " Consider setting `delta=0.0` in your query."
                raise ValueError(msg)

        chain = self.query_space or self.accountant.input_space
        query = Query(
            chain=chain,
            output_measure=self.accountant.output_measure,
            d_in=self.d_in,
            d_out=d_query,
            context=self,
        )

        # return a LazyFrameQuery when dealing with Polars data, to better mimic the Polars API
        if chain[0].type == "LazyFrameDomain":
            from opendp.domains import _lazyframe_from_domain

            # creates an empty lazyframe to hold the query plan
            polars_plan = _lazyframe_from_domain(self.accountant.input_domain)
            return LazyFrameQuery(polars_plan, query)

        return query

    def current_privacy_loss(self):
        """When the query losses are static, returns the list of consumed privacy losses,
        otherwise returns the current privacy loss of the queryable."""
        if isinstance(self.queryable, OdometerQueryable):
            return self.queryable.privacy_loss(self.d_in)
        return self.d_mids_consumed.copy()

    def remaining_privacy_loss(self):
        """Returns the remaining privacy loss."""
        if isinstance(self.queryable, OdometerQueryable):
            if self.d_out is None:
                raise ValueError("The privacy loss is unbounded.")

            def _sub(a, b):
                if (
                    isinstance(a, tuple)
                    and isinstance(b, tuple)
                    and len(a) == len(b) == 2
                ):
                    return a[0] - b[0], a[1] - b[1]
                else:
                    return a - b

            return _sub(self.d_out, self.queryable.privacy_loss(self.d_in))

        return None if self.d_mids is None else self.d_mids.copy()


Chain = Union[
    tuple[Domain, Metric], Transformation, Measurement, Odometer, "PartialChain"
]


class Query(object):
    """Initializes the query with the given chain and output measure.

    It is more convenient to use the ``context.query()`` constructor than this one.
    However, this can be used stand-alone to help build a transformation/measurement that is not part of a context.

    :param chain: an initial metric space (tuple of domain and metric) or transformation
    :param output_measure: how privacy will be measured on the output of the query
    :param d_in: an upper bound on the distance between adjacent datasets
    :param d_out: an upper bound on the overall privacy loss
    :param context: if specified, then when the query is released, the chain will be submitted to this context
    :param _wrap_release: for internal use only
    """

    _chain: Chain
    """The current chain of transformations and measurements."""
    _output_measure: Measure
    """The output measure of the query."""
    _context: Optional["Context"]
    """The context that the query is part of. ``query.release()`` submits ``_chain`` to ``_context``."""
    _wrap_release: Optional[Callable[[Any], Any]]
    """For internal use. A function that wraps the release of the query. 
    Used to wrap the response of compositor/odometer queries in another ``Context``."""

    def __init__(
        self,
        chain: Chain,
        output_measure: Measure,
        d_in: Optional[Union[float, Sequence[Bound]]] = None,
        d_out: Optional[Union[float, tuple[float, float]]] = None,
        context: Optional["Context"] = None,
        _wrap_release=None,
    ) -> None:
        self._chain = chain
        self._output_measure = output_measure
        self._d_in = d_in
        self._d_out = d_out
        self._context = context
        self._wrap_release = _wrap_release

    def __repr__(self) -> str:
        context = ""
        if self._context is not None:
            context = f""",\n    context        = {indent(repr(self._context))}"""

        return f"""Query(
    chain          = {indent(repr(self._chain))},
    output_measure = {self._output_measure},
    d_in           = {self._d_in},
    d_out          = {self._d_out}{context})"""

    def __getattr__(self, name: str) -> Callable[..., "Query"]:
        """Creates a new query by applying a transformation or measurement to the current chain."""
        if name == "canonical_noise":
            return self._canonical_noise

        if name not in constructors:
            raise AttributeError(
                f"Unrecognized constructor: '{name}'. Did you mean to '.release()' first?"
            )  # pragma: no cover

        def make(*args, **kwargs) -> "Query":
            """Wraps the ``make_{name}`` constructor to allow one optional parameter and chains it to the current query.

            This function will be called when the user calls ``query.{name}(...)``.

            :param args: arguments for the constructor function
            :param kwargs: keyword arguments for the constructor function
            """
            constructor, is_partial = constructors[name]

            # determine how many parameters are missing
            param_diff = len(args)
            for param in signature(constructor).parameters.values():
                if param.name in kwargs:
                    continue
                if param.default is not param.empty:
                    break
                param_diff -= 1

            if param_diff == -1 and not isinstance(self._chain, PartialChain):
                constructor = PartialChain.wrap(constructor)
            elif param_diff < 0:
                raise ValueError(f"{name} is missing {-param_diff} parameter(s).")
            elif param_diff > 0:
                raise ValueError(
                    f"{name} has {param_diff} parameter(s) too many."
                )  # pragma: no cover

            new_chain = constructor(*args, **kwargs)
            if is_partial or not isinstance(self._chain, tuple):
                new_chain = self._chain >> new_chain

            return self.new_with(chain=new_chain)

        return make

    def _canonical_noise(self, binomial_size: Optional[int] = None):
        """Make a measurement that adds noise from the canonical noise distribution.

        :param binomial_size: (approximate) total number of records if dataset consists of bernoulli draws
        """
        # left as private method so that it behaves like other constructors in IDEs
        # this is getting a special wrapper because
        # - its d_in argument is the stability of the prior transformation
        # - its d_out argument is already stored in the chain
        # - it has a special postprocessor
        from opendp.measurements import then_canonical_noise
        from opendp._internal import _new_pure_function
        from opendp.extras.numpy.canonical import BinomialCND

        def then(d_in, d_out):
            m_noise = then_canonical_noise(d_in, d_out)
            if binomial_size is not None:
                m_noise = m_noise >> _new_pure_function(
                    lambda x: BinomialCND(x, d_in, d_out, binomial_size),
                    TO="ExtrinsicObject",
                )
            return m_noise

        if isinstance(self._chain, tuple):
            d_mid = self._d_in
        elif isinstance(self._chain, Transformation):
            d_mid = self._chain.map(self._d_in)
        elif isinstance(self._chain, PartialChain):
            raise ValueError(
                "Canonical noise requires all arguments in the input query to be specified."
            )
        else:
            raise ValueError(  # pragma: no cover
                f"Canonical noise expects a metric space or transformation as the prior query, found {self._chain}"
            )

        return self.new_with(chain=self._chain >> then(d_mid, self._d_out))

    def new_with(self, *, chain: Chain, wrap_release=None) -> "Query":
        """Convenience constructor that creates a new query with a different chain.

        :param chain: the prior query. Either a metric space or transformation
        :param wrap_release: a function to apply to apply to releases
        """
        return Query(
            chain=chain,
            output_measure=self._output_measure,
            d_in=self._d_in,
            d_out=self._d_out,
            context=self._context,  # type: ignore[arg-type]
            _wrap_release=wrap_release or self._wrap_release,
        )

    def __dir__(self):
        """Returns the list of available constructors. Used by Python's error suggestion mechanism.
        Without this, none of the transformations or measument methods are listed.
        """
        return super().__dir__() + list(constructors.keys())  # type: ignore[operator]

    def resolve(
        self,
        allow_transformations: bool = False,
        bounds: Optional[tuple[float, float]] = None,
        T: Optional[Union[Type[float], Type[int]]] = None,
    ) -> Union[Transformation, Measurement, Odometer]:
        """Resolve the query into a transformation, measurement or odometer.

        :param allow_transformations: If true, allow the response to be a transformation.
        :param bounds: Search for a potentially missing parameter of interest within these bounds.
        :param T: The type of the parameter to search for. Either ``float`` or ``int``.
        """
        # resolve a partial chain into a measurement, by fixing the input and output distances
        if isinstance(self._chain, PartialChain):
            assert self._d_in is not None
            assert self._d_out is not None
            chain = self._chain.fix(
                self._d_in, self._d_out, self._output_measure, bounds=bounds, T=T
            )
        else:
            chain = self._chain
        if not allow_transformations and isinstance(chain, Transformation):
            raise ValueError("Query is not yet a measurement or odometer.")
        return _cast_measure(chain, self._output_measure, self._d_out)

    def release(
        self,
        data=None,
        bounds: Optional[tuple[float, float]] = None,
        T: Optional[Union[Type[float], Type[int]]] = None,
    ):
        """Release the query. The query must be part of a context.

        :param data: The data to be analyzed. If not specified, the query must be part of a context.
        :param bounds: Search for a potentially missing parameter of interest within these bounds.
        :param T: The type of the parameter to search for. Either ``float`` or ``int``.
        """
        if self._context is not None and data is not None:
            raise ValueError("Cannot specify data when the query is part of a context.")

        measurement = self.resolve(bounds=bounds, T=T)

        if self._context is not None:
            answer = self._context(measurement)
        elif data is not None:
            answer = measurement(data)
        else:
            raise ValueError("Cannot release query without data or context.")

        if self._wrap_release:
            answer = self._wrap_release(answer)

        return answer

    def param(
        self,
        allow_transformations=False,
        bounds: Optional[tuple[float, float]] = None,
        T: Optional[Union[Type[float], Type[int]]] = None,
    ):
        """Returns the discovered parameter, if there is one.

        :param allow_transformations: If true, allow the response to be a transformation.
        :param bounds: Search for a potentially missing parameter of interest within these bounds.
        :param T: The type of the parameter to search for. Either ``float`` or ``int``.
        """
        resolved = self.resolve(
            allow_transformations=allow_transformations, bounds=bounds, T=T
        )
        return getattr(resolved, "param", None)

    def compositor(
        self,
        split_evenly_over: Optional[int] = None,
        split_by_weights: Optional[Sequence[float]] = None,
        d_out: Optional[Union[float, tuple[float, float]]] = None,
        output_measure: Optional[Measure] = None,
        alpha: Optional[float] = None,
    ) -> "Query":
        """Constructs a new context containing a sequential compositor with the given weights.

        ``split_evenly_over`` and ``split_by_weights`` are mutually exclusive.

        :param split_evenly_over: The number of parts to evenly distribute the privacy loss
        :param split_by_weights: A list of weights for each intermediate privacy loss
        :param d_out: Optional upper bound on privacy loss.
        :param output_measure: Optional method of accounting to be used by this compositor. Defaults to same.
        :param alpha: Optional parameter to split delta between zCDP conversion and δ-approximate in approx-ZCDP
        """

        d_out = d_out or self._d_out

        if d_out is None:
            raise ValueError(
                "`d_out` is unknown. Please specify it in the query."
            )  # pragma: no cover

        if output_measure is not None:
            d_out = _translate_measure_distance(
                d_out, self._output_measure, output_measure, alpha
            )

        def _compositor(chain: Union[tuple[Domain, Metric], Transformation], d_in):
            if isinstance(chain, tuple):
                input_domain, input_metric = chain
            elif isinstance(chain, Transformation):
                input_domain, input_metric = chain.output_domain, chain.output_metric
                d_in = chain.map(d_in)

            privacy_unit = input_metric, d_in
            privacy_loss = output_measure or self._output_measure, d_out

            accountant, d_mids, _ = _normalize_compositor(
                input_domain,
                privacy_unit,
                privacy_loss,
                split_evenly_over,
                split_by_weights,
            )
            if isinstance(chain, Transformation):
                accountant = chain >> accountant

            def _wrap_release(queryable):
                return Context(
                    accountant=accountant,
                    queryable=queryable,
                    d_in=d_in,
                    d_mids=d_mids,
                    d_out=d_out,
                    query_space=(input_domain, input_metric),
                )

            return self.new_with(chain=accountant, wrap_release=_wrap_release)

        return self._compose_context(_compositor)

    def _compose_context(self, compositor):
        """Helper function for composition in a context."""
        if isinstance(self._chain, PartialChain):
            # TODO: Can we exercise this?
            return PartialChain(
                lambda x: compositor(self._chain(x), self._d_in)
            )  # pragma: no cover
        else:
            return compositor(self._chain, self._d_in)


class PartialChain(object):
    """A partial chain is a transformation or measurement that is missing one numeric parameter.

    The parameter can be solved for by calling the fix method,
    which returns the closest transformation or measurement that satisfies the given stability or privacy constraint.
    """

    partial: Callable[[float], Union[Transformation, Measurement]]
    """The partial transformation or measurement."""

    def __init__(self, f, *args, **kwargs):
        self.partial = partial(f, *args, **kwargs)

    def __call__(self, v):
        """Returns the transformation or measurement with the given parameter."""
        return self.partial(v)

    def fix(
        self,
        d_in: Union[float, Sequence[Bound]],
        d_out: Union[float, tuple[float, float]],
        output_measure: Optional[Measure] = None,
        bounds: Optional[tuple[float, float]] = None,
        T=None,
    ):
        """Returns the closest transformation or measurement that satisfies the given stability or privacy constraint.

        The discovered parameter is assigned to the param attribute of the returned transformation or measurement.

        :param d_in: Upper bound on the distance between adjacent input datasets.
        :param d_out: Upper bound on the distance between adjacent output distributions.
        :param output_measure: How to measure distances between output distributions.
        :param bounds: The bounds for the parameter search.
        :param T: The type of the parameter to search for.
        """
        # When the output measure corresponds to approx-DP, only optimize the epsilon parameter.
        # The delta parameter should be fixed in _cast_measure, and if not, then the search will be impossible here anyways.
        if (
            output_measure is not None
            and isinstance(output_measure.type, RuntimeType)
            and output_measure.type.origin == "Approximate"
        ):

            def _predicate(param):
                meas = _cast_measure(self(param), output_measure, d_out)
                return meas.map(d_in)[0] <= d_out[0]  # type: ignore[index]

        else:

            def _predicate(param):
                meas = _cast_measure(self(param), output_measure, d_out)
                return meas.check(d_in, d_out)

        param = binary_search(_predicate, bounds=bounds, T=T)
        chain = self.partial(param)
        chain.param = param
        return chain

    def __rshift__(
        self, other: Union[Transformation, Measurement, _PartialConstructor]
    ):
        # partials may be chained with other transformations or measurements to form a new partial
        if isinstance(other, (Transformation, Measurement, _PartialConstructor)):
            return PartialChain(lambda x: self(x) >> other)

        raise ValueError(
            "At most one parameter may be missing at a time"
        )  # pragma: no cover

    def __rrshift__(
        self, other: Union[tuple[Domain, Metric], Transformation, Measurement]
    ):
        if isinstance(other, (tuple, Transformation, Measurement)):
            return PartialChain(lambda x: other >> self(x))

        raise ValueError(
            "At most one parameter may be missing at a time"
        )  # pragma: no cover

    @classmethod
    def wrap(cls, f):
        """Wraps a constructor for a transformation or measurement to return a partial chain instead.

        :param f: function to wrap
        """

        def _inner(*args, **kwargs):
            return cls(f, *args, **kwargs)

        return _inner


def _normalize_compositor(
    domain: Domain,
    privacy_unit: tuple[Metric, Union[float, Sequence[Bound]]],
    privacy_loss: tuple[Measure, Union[float, tuple[float, float]]],
    split_evenly_over: Optional[int] = None,
    split_by_weights: Optional[Sequence[float]] = None,
) -> tuple[
    Union[Measurement, Odometer],
    Optional[Union[list[float], list[tuple[float, float]]]],
    Optional[Any],
]:
    """Constructs a sequential composition measurement
    where the ``d_mids`` are proportional to the weights.

    ``split_evenly_over`` and ``split_by_weights`` are mutually exclusive.

    * composition: returns a measurement, d_mids, and None
    * filter: returns a measurement, None, and d_out
    * adaptive: returns a measurement, None, and None

    :param domain: the domain of the data
    :param privacy_unit: a tuple of the input metric and the data distance (``d_in``)
    :param privacy_loss: a tuple of the output measure and the privacy parameter (``d_out``)
    :param split_evenly_over: The number of parts to evenly distribute the privacy loss
    :param split_by_weights: A list of weights for each intermediate privacy loss
    :return: a tuple of the measurement, d_mids, and d_out
    """
    input_metric, d_in = privacy_unit
    output_measure, d_out = privacy_loss

    if split_evenly_over is not None and split_by_weights is not None:
        raise ValueError(
            "Cannot specify both `split_evenly_over` and `split_by_weights`"
        )  # pragma: no cover

    if split_evenly_over is not None:
        weights = [d_out] * split_evenly_over
    elif split_by_weights is not None:
        # TODO: Wrap d_out in an object that defines __mul__,
        # and fix the typing on the signature.
        if isinstance(d_out, float):
            weights = [d_out * w for w in split_by_weights]
        else:
            weights = [(d_out[0] * w, d_out[1] * w) for w in split_by_weights]
    else:
        odometer = make_fully_adaptive_composition(
            input_domain=domain,
            input_metric=input_metric,
            output_measure=output_measure,
        )

        inf = float("inf")
        if d_out in {inf, (inf, inf), (inf, 1.0)}:
            return odometer, None, None

        return make_privacy_filter(odometer, d_in, d_out), None, d_out

    def _mul(dist, scale: float):
        if isinstance(dist, tuple):
            return dist[0] * scale, dist[1] * scale
        else:
            return dist * scale

    def _scale_weights(scale: float, weights):
        return [_mul(w, scale) for w in weights]

    def _scale_ac(scale: float):
        return make_adaptive_composition(
            input_domain=domain,
            input_metric=input_metric,
            output_measure=output_measure,
            d_in=d_in,
            d_mids=_scale_weights(scale, weights),
        )

    scale = binary_search_param(_scale_ac, d_in=d_in, d_out=d_out, T=float)

    # return the accountant and d_mids
    return _scale_ac(scale), _scale_weights(scale, weights), None


def _cast_measure(chain, to_measure: Optional[Measure] = None, d_to=None):
    """Casts the output measure of a given ``chain`` to ``to_measure``.

    If provided, ``d_to`` is the privacy loss wrt the new measure.
    """
    if to_measure is None or chain.output_measure == to_measure:
        return chain

    from_to = str(chain.output_measure.type), str(to_measure.type)

    if from_to == ("MaxDivergence", "Approximate<MaxDivergence>"):
        return make_approximate(chain)

    if from_to == (
        "ZeroConcentratedDivergence",
        "Approximate<ZeroConcentratedDivergence>",
    ):
        return make_approximate(chain)

    if from_to == ("MaxDivergence", "ZeroConcentratedDivergence"):
        return make_pureDP_to_zCDP(chain)

    if from_to == (
        "ZeroConcentratedDivergence",
        "Approximate<MaxDivergence>",
    ) or from_to == (
        "Approximate<ZeroConcentratedDivergence>",
        "Approximate<MaxDivergence>",
    ):
        return make_fix_delta(make_zCDP_to_approxDP(chain), d_to[1])

    raise ValueError(
        f"Unable to cast measure from {from_to[0]} to {from_to[1]}"
    )  # pragma: no cover


def _translate_measure_distance(
    d_from, from_measure: Measure, to_measure: Measure, alpha: Optional[float] = None
) -> Union[float, tuple[float, float]]:
    """Translate a privacy loss ``d_from`` from ``from_measure`` to ``to_measure``.

    >>> _translate_measure_distance(1, dp.max_divergence(), dp.max_divergence())
    1
    >>> _translate_measure_distance(1, dp.max_divergence(), dp.approximate(dp.max_divergence()))
    (1, 0.0)
    >>> _translate_measure_distance((1.5, 5e-07), dp.approximate(dp.max_divergence()), dp.zero_concentrated_divergence())
    0.0489...
    >>> _translate_measure_distance(0.05, dp.zero_concentrated_divergence(), dp.max_divergence())
    0.316...
    """
    if from_measure == to_measure:
        return d_from

    from_to = str(from_measure.type), str(to_measure.type)

    constant = 1.0  # the choice of constant doesn't matter

    if from_to == ("MaxDivergence", "Approximate<MaxDivergence>"):
        return (d_from, 0.0)

    if from_to == ("ZeroConcentratedDivergence", "MaxDivergence"):
        space = atom_domain(T=float, nan=False), absolute_distance(T=float)
        scale = binary_search_param(
            lambda scale: make_pureDP_to_zCDP(make_laplace(*space, scale)),
            d_in=constant,
            d_out=d_from,
            T=float,
        )
        return make_laplace(*space, scale).map(constant)

    if from_to == (
        "Approximate<MaxDivergence>",
        "ZeroConcentratedDivergence",
    ):

        def _caster(measurement):
            return make_fix_delta(make_zCDP_to_approxDP(measurement), delta=d_from[1])

        space = atom_domain(T=int), absolute_distance(T=float)
        scale = binary_search_param(
            lambda scale: _caster(make_gaussian(*space, scale)),
            d_in=constant,
            d_out=d_from,
            T=float,
        )
        return make_gaussian(*space, scale).map(constant)

    if from_to == (
        "Approximate<MaxDivergence>",
        "Approximate<ZeroConcentratedDivergence>",
    ):
        epsilon, delta = d_from
        if alpha is None or not (0 <= alpha < 1):
            raise ValueError(f"alpha ({alpha}) must be in [0, 1)")  # pragma: no cover
        delta_zCDP, delta_inf = delta * (1 - alpha), delta * alpha

        rho = _translate_measure_distance(
            (epsilon, delta_zCDP), from_measure, zero_concentrated_divergence()
        )
        return rho, delta_inf  # type: ignore[return-value]

    raise ValueError(
        f"Unable to translate distance from {from_to[0]} to {from_to[1]}"
    )  # pragma: no cover



# Exhaustively checks each 32-bit float between [0, inf) to find the greatest error (in ulps) of the erfc function.
# The result from this check is that the erfc function in statrs errs by at most 1 f32 ulp.

# First disagreement is at 0.5.
# Execution is slowest at inputs around 15, before switching to the next approximating curve.

# To use all CPUs, floats are sharded modulo the number of CPUs (less two).
# It may be necessary to restart this program at a later float to free memory.

import struct
import multiprocessing

# pip install gmpy2
import gmpy2
from opendp._data import erfc

# specifically check max ulp distance from a conservative upper bound
gmpy2.get_context().round = gmpy2.RoundUp


def floatToBits(f):
    s = struct.pack(">f", f)
    return struct.unpack(">l", s)[0]


def bitsToFloat(b):
    s = struct.pack(">l", b)
    return struct.unpack(">f", s)[0]


def worker(offset, step):
    max_err = 0
    print(f"running {offset}")
    # iterate through all 32-bit floats
    for bits in range(floatToBits(0.0), floatToBits(float("inf")), step):
        bits += offset
        if offset == 0 and bits > 0 and (bits // step) % 10_000 == 0:
            prop_done = bits / floatToBits(float("inf"))
            print(
                f"{prop_done:.2%} done, with max discovered f32 ulp error of: {max_err}. Currently at: {bitsToFloat(bits)}"
            )

        f32 = bitsToFloat(bits)
        f32_ulp_err = abs(floatToBits(erfc(f32)) - floatToBits(gmpy2.erfc(f32)))
        max_err = max(max_err, f32_ulp_err)
    print(max_err)


if __name__ == "__main__":
    n_cpus = multiprocessing.cpu_count() - 2
    processes = []
    for cpu in range(n_cpus):
        p = multiprocessing.Process(target=worker, args=(cpu, n_cpus))
        p.start()
        processes.append(p)

    [p.join() for p in processes]



import json
import re

from opendp.extras.polars import Margin
import pytest

import opendp.prelude as dp
from opendp._lib import import_optional_dependency

atom = dp.atom_domain(bounds=(0., 10.), nan=False)
input_space = dp.vector_domain(atom, size=10), dp.symmetric_distance()
chained = input_space >> dp.t.then_mean() >> dp.m.then_laplace(scale=0.5)


@pytest.mark.parametrize(
    "dp_obj",
    [
        # Python objects:
        ('nested', ('tuple', ('containing', ('domain', (atom,))))),
        {'dict key': atom},
        input_space,
        # Domains:
        atom,
        dp.categorical_domain(['A', 'B', 'C']),
        dp.series_domain('A', atom),
        dp.lazyframe_domain([dp.series_domain('A', atom)]),
        dp.wild_expr_domain([]),
        # Metrics:
        dp.absolute_distance("int"),
        dp.change_one_distance(),
        dp.linf_distance("float", True),
        dp.user_distance("user_distance"),
        # Measures:
        dp.m.max_divergence(),
        dp.m.approximate(dp.m.max_divergence()),
        dp.m.user_divergence("user_divergence"),
        # Measurements:
        dp.m.make_gaussian(atom, dp.absolute_distance(float), 1),
        dp.m.then_gaussian(1),
        # Compositions:
        chained,
        dp.c.make_population_amplification(chained, population_size=100),
    ],
    ids=lambda arg: str(arg)
)
def test_serializable_equal(dp_obj):
    serialized = dp.serialize(dp_obj)
    deserialized = dp.deserialize(serialized)
    # We don't want to define __eq__ just for the sake of testing,
    # so check the serializations before and after.
    # (We should remember that if the first serialization
    #  dropped some detail, this test wouldn't catch it.)
    assert serialized == dp.serialize(deserialized)


@pytest.mark.parametrize(
    "dp_obj",
    [
        dp.user_domain("trivial_user_domain", lambda: True),
        dp.m.new_privacy_profile(lambda x: x),
    ],
    ids=lambda arg: str(arg)
)
def test_not_ever_serializable(dp_obj):
    with pytest.raises(Exception, match=r"OpenDP JSON Encoder does not handle"):
        dp.serialize(dp_obj)


@pytest.mark.parametrize(
    "dp_obj",
    [
        {('tuple', 'key'): 'value'},
    ],
    ids=lambda arg: str(arg)
)
def test_not_json_serializable(dp_obj):
    with pytest.raises(Exception, match=r"keys must be str, int, float, bool or None, not tuple"):
        dp.serialize(dp_obj)


def test_version_mismatch_warning():
    bad_serialized = json.dumps({
        "__function__": "atom_domain",
        "__module__": "domains",
        "__kwargs__": {
            "bounds": {"__tuple__": [0, 10]}, 
            "nan": False, 
            "T": "i32"
        },
        "__version__": "bad-version"
    })
    with pytest.warns(UserWarning, match=re.escape('(bad-version) != this version')):
        dp.deserialize(bad_serialized)


def test_deserialize_polars_plan_error():
    pl = pytest.importorskip('polars')
    context = dp.Context.compositor(
        data=[1, 2, 3], # It will expect a LazyFrame here.
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=1,
    )
    with pytest.raises(ValueError, match=r"'data' of context must be a LazyFrame"):
        context.deserialize_polars_plan(pl.LazyFrame({}).serialize())


# Would normally put the conditional inside the test,
# but since we need polars at test collection time,
# in this case it needs to wrap the tests.
pl = import_optional_dependency('polars', raise_error=False)
if pl is not None:
    # member() will warn if instance is not even of the carrier type,
    # but that behavior is tested elsewhere, and can be ignored here.
    @pytest.mark.filterwarnings("ignore::UserWarning")
    @pytest.mark.parametrize(
        "dp_domain,in_value,out_value",
        [
            (atom, 10, 100),
            # TODO: Might not be specifying categorical values correctly, 
            # but shouldn't error, regardless.
            # https://github.com/opendp/opendp/issues/2264
            # (dp.categorical_domain(['A', 'B', 'C']),
            #  pl.lit("A", dtype=pl.Categorical),
            #  pl.lit("Z", dtype=pl.Categorical)
            # ),
            (
                dp.series_domain('name', atom),
                pl.Series("name", [1.0, 2.0, 3.0]),
                pl.Series("name", ['a', 'b', 'c'])
            ),
            (
                dp.lazyframe_domain([dp.series_domain('A', atom)]),
                pl.LazyFrame({'A': [1.0, 2.0, 3.0]}),
                pl.LazyFrame({'A': ['a', 'b', 'c']})
            )
        ],
        ids=lambda arg: str(arg)
    )
    def test_serializable_domain(dp_domain, in_value, out_value):
        assert dp_domain.member(in_value)
        assert not dp_domain.member(out_value)

        serialized = dp.serialize(dp_domain)
        deserialized = dp.deserialize(serialized)

        assert deserialized.member(in_value)
        assert not deserialized.member(out_value)

    @pytest.mark.parametrize(
        "dp_measurement,value,output_type",
        [
            (dp.m.make_gaussian(atom, dp.absolute_distance(float), 1), 0, float),
        ],
        ids=lambda arg: str(arg)
    )
    def test_serializable_measurement(dp_measurement, value, output_type):
        assert isinstance(dp_measurement(value), output_type)
        


    lf = pl.LazyFrame(schema={"A": pl.Int32, "B": pl.String})
    lf_domain = dp.lazyframe_domain([
        dp.series_domain("A", dp.atom_domain(T="i32")), 
        dp.series_domain("B", dp.atom_domain(T=str))
    ])
    lf_domain_with_margin = dp.with_margin(lf_domain, Margin(by=[], max_length=1000))

    context = dp.Context.compositor(
        data=pl.LazyFrame({"age": [1, 2, 3]}),
        privacy_unit=dp.unit_of(contributions=1),
        privacy_loss=dp.loss_of(epsilon=1.0),
        split_evenly_over=10,
    )
    query = context.query().select(dp.len())

    @pytest.mark.parametrize(
        "dp_obj",
        [
            lf_domain,
            lf_domain_with_margin,
            dp.m.make_private_lazyframe(
                lf_domain_with_margin,
                dp.symmetric_distance(),
                dp.max_divergence(),
                lf.select([dp.len(), pl.col("A").dp.sum((0, 1))]),
                global_scale=1.0
            ),
            dp.m.make_private_expr(
                dp.wild_expr_domain([], dp.polars.Margin(by=[])),
                dp.l01inf_distance(dp.symmetric_distance()),
                dp.max_divergence(),
                dp.len(scale=1.0)
            ),
        ],
        ids=lambda arg: str(arg)
    )
    def test_serializable_polars(dp_obj):
        serialized = dp.serialize(dp_obj)
        deserialized = dp.deserialize(serialized)
        assert serialized == dp.serialize(deserialized)


    @pytest.mark.parametrize(
        "dp_obj",
        [
            context,
            dp.Queryable('value', 'query_type'),
            query
        ],
        ids=lambda arg: str(arg)
    )
    def test_not_currently_serializable(dp_obj):
        with pytest.raises(Exception, match=r"OpenDP JSON Encoder does not handle"):
            dp.serialize(dp_obj)



# type: ignore
def make_gaussian(
    input_domain: DI,
    input_metric: MI,
    scale: f64,
    k: Option[i32],
) -> Measurement[DI, DI_Carrier, MI, MO]:
    return DiscreteGaussian(scale, k).make_noise((input_domain, input_metric))



# type: ignore
class IntExpFamily:
    def make_noise(
        self, input_space: tuple[VectorDomain[AtomDomain[T]], LpDistance[P, QI]]
    ) -> Measurement[VectorDomain[AtomDomain[T]], T, LpDistance[P, QI], MO]:
        distribution = ZExpFamily(
            scale=integerize_scale(self.scale, 0)
        )  # |\label{line:dist}|

        t_int = make_int_to_bigint(input_space)
        m_noise = distribution.make_noise(t_int.output_space())
        return t_int >> m_noise >> then_saturating_cast()



# type: ignore
def make_float_to_bigint_threshold(
    input_space: tuple[
        MapDomain[AtomDomain[TK], MapDomain[TV]], L0PInfDistance[P, AbsoluteDistance[QI]]
    ],
    threshold: TV,
    k: i32,
) -> Transformation[
    MapDomain[AtomDomain[TK], AtomDomain[TV]],
    MapDomain[AtomDomain[TK], AtomDomain[IBig]],
    L0PInfDistance[P, AbsoluteDistance[QI]],
    L0PInfDistance[P, AbsoluteDistance[RBig]],
]:
    input_domain, input_metric = input_space
    if input_domain.value_domain.nan():
        raise "input_domain hashmap values may not contain NaN elements"

    r_threshold = RBig.try_from(threshold)

    min_k = get_min_k(TV)
    if k < min_k:  # |\label{line:check-k}|
        raise f"k ({k}) must not be smaller than {min_k}"

    def value_function(val):  # |\label{line:value-function}|
        try:  # |\label{line:try-from}|
            val = RBig.try_from(val)
        except Exception:
            val = RBig.ZERO

        return find_nearest_multiple_of_2k(val, k)  # |\label{line:find-nearest}|

    def stability_map(d_in):
        l0, lp, li = d_in

        r_lp = RBig.try_from(lp)
        r_lp_round = get_rounding_distance(k, usize.from_(l0), P)
        r_lp = x_mul_2k(r_lp + r_lp_round, -k)  # |\label{line:lp-x-mul-2k}|

        r_li = RBig.try_from(li)
        if r_li > x_mul_2k(r_threshold, -k): # |\label{line:check-li}|
            raise f"li ({li}) must not be larger than threshold ({threshold})"
        r_li_round = get_rounding_distance(k, 1, P)
        r_li = x_mul_2k(r_li + r_li_round, -k)  # |\label{line:li-x-mul-2k}|

        return l0, r_lp, r_li

    return Transformation.new(
        input_domain,
        MapDomain(  # |\label{line:output-domain}|
            key_domain=input_domain.key_domain,
            value_domain=AtomDomain.default(IBig),
        ),
        Function.new(lambda x: {k: value_function(v) for k, v in x.items()}),
        input_metric,
        L0PI.default(),
        StabilityMap.new_fallible(stability_map),
    )



# type: ignore
# analogous to impl MakeNoise<VectorDomain<AtomDomain<IBig>>, MI, MO> for RV in Rust
class RV:
    def make_noise(self, input_space) -> Measurement[VectorDomain[AtomDomain[IBig]], Vec[IBig], MI, MO]:
        input_domain, input_metric = input_space
        return Measurement.new(
            input_domain,
            Function.new_fallible(
                lambda x: [self.sample(x_i) for x_i in x]), # |\label{line:sample}|
            input_metric,
            MO.default(),
            self.noise_privacy_map(),  # |\label{line:privacy-map}|
        ) 



# type: ignore
def x_mul_2k(x: RBig, k: i32) -> RBig:
    num, den = x.into_parts()  # |\label{line:into-parts}|
    if k < 0:
        den <<= cast(-k, usize)
    else:
        num <<= cast(k, usize)

    return RBig.from_parts(num, den)



# Auto-generated. Do not edit!
'''
The ``combinators`` module provides functions for combining transformations and measurements.
For more context, see :ref:`combinators in the User Guide <combinators-user-guide>`.

For convenience, all the functions of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The methods of this module will then be accessible at ``dp.c``.
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401
from opendp.core import *
from opendp.domains import *
from opendp.metrics import *
from opendp.measures import *
__all__ = [
    "make_adaptive_composition",
    "make_approximate",
    "make_basic_composition",
    "make_chain_mt",
    "make_chain_pm",
    "make_chain_tt",
    "make_composition",
    "make_fix_delta",
    "make_fixed_approxDP_to_approxDP",
    "make_fully_adaptive_composition",
    "make_population_amplification",
    "make_privacy_filter",
    "make_pureDP_to_zCDP",
    "make_select_private_candidate",
    "make_sequential_composition",
    "make_zCDP_to_approxDP",
    "then_adaptive_composition",
    "then_fully_adaptive_composition",
    "then_sequential_composition"
]


def make_adaptive_composition(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    d_in,
    d_mids
) -> Measurement:
    r"""Construct a Measurement that when invoked,
    returns a queryable that interactively composes measurements.

    **Composition Properties**

    * sequential: all measurements are applied to the same dataset
    * basic: the composition is the linear sum of the privacy usage of each query
    * interactive: mechanisms can be specified based on answers to previous queries
    * compositor: all privacy parameters specified up-front

    If the privacy measure supports concurrency,
    this compositor allows you to spawn multiple interactive mechanisms
    and interleave your queries amongst them.


    Required features: `contrib`

    [make_adaptive_composition in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_adaptive_composition.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `Queryable<Measurement<DI, MI, MO, TO>, TO>`

    .. end-markdown

    :param input_domain: indicates the space of valid input datasets
    :type input_domain: Domain
    :param input_metric: how distances are measured between members of the input domain
    :type input_metric: Metric
    :param output_measure: how privacy is measured
    :type output_measure: Measure
    :param d_in: maximum distance between adjacent input datasets
    :param d_mids: maximum privacy expenditure of each query
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    QO = get_distance_type(output_measure) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_d_in = py_to_c(d_in, c_type=AnyObjectPtr, type_name=get_distance_type(input_metric))
    c_d_mids = py_to_c(d_mids, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=[QO]))

    # Call library function.
    lib_function = lib.opendp_combinators__make_adaptive_composition
    lib_function.argtypes = [Domain, Metric, Measure, AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_d_in, c_d_mids), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_adaptive_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'd_in': d_in, 'd_mids': d_mids
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_adaptive_composition(
    output_measure: Measure,
    d_in,
    d_mids
):  
    r"""partial constructor of make_adaptive_composition

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.combinators.make_adaptive_composition`

    :param output_measure: how privacy is measured
    :type output_measure: Measure
    :param d_in: maximum distance between adjacent input datasets
    :param d_mids: maximum privacy expenditure of each query
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_adaptive_composition(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        d_in=d_in,
        d_mids=d_mids))
    output.__opendp_dict__ = {
            '__function__': 'then_adaptive_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'output_measure': output_measure, 'd_in': d_in, 'd_mids': d_mids
            },
        }
    return output



def make_approximate(
    measurement: Measurement
) -> Measurement:
    r"""Constructs a new output measurement where the output measure
    is δ-approximate, where δ=0.


    Required features: `contrib`

    [make_approximate in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_approximate.html)

    .. end-markdown

    :param measurement: a measurement with a privacy measure to be casted
    :type measurement: Measurement
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")

    # Call library function.
    lib_function = lib.opendp_combinators__make_approximate
    lib_function.argtypes = [Measurement]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_approximate',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


@deprecated(version="0.14.0", reason="This function has been renamed, use `make_composition` instead.")
def make_basic_composition(
    measurements
) -> Measurement:
    r"""Construct the DP composition \[`measurement0`, `measurement1`, ...\].
    Returns a Measurement that when invoked, computes `[measurement0(x), measurement1(x), ...]`

    All metrics and domains must be equivalent.

    **Composition Properties**

    * sequential: all measurements are applied to the same dataset
    * basic: the composition is the linear sum of the privacy usage of each query
    * noninteractive: all mechanisms specified up-front (but each can be interactive)
    * compositor: all privacy parameters specified up-front (via the map)


    Required features: `contrib`

    .. end-markdown

    :param measurements: A vector of Measurements to compose.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurements = py_to_c(measurements, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["AnyMeasurementPtr"]))

    # Call library function.
    lib_function = lib.opendp_combinators__make_basic_composition
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurements), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_basic_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurements': measurements
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_chain_mt(
    measurement1: Measurement,
    transformation0: Transformation
) -> Measurement:
    r"""Construct the functional composition (`measurement1` ○ `transformation0`).
    Returns a Measurement that when invoked, computes `measurement1(transformation0(x))`.


    Required features: `contrib`

    [make_chain_mt in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_chain_mt.html)

    .. end-markdown

    :param measurement1: outer mechanism
    :type measurement1: Measurement
    :param transformation0: inner transformation
    :type transformation0: Transformation
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement1 = py_to_c(measurement1, c_type=Measurement, type_name=None)
    c_transformation0 = py_to_c(transformation0, c_type=Transformation, type_name=None)

    # Call library function.
    lib_function = lib.opendp_combinators__make_chain_mt
    lib_function.argtypes = [Measurement, Transformation]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement1, c_transformation0), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_chain_mt',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement1': measurement1, 'transformation0': transformation0
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_chain_pm(
    postprocess1: Function,
    measurement0: Measurement
) -> Measurement:
    r"""Construct the functional composition (`postprocess1` ○ `measurement0`).
    Returns a Measurement that when invoked, computes `postprocess1(measurement0(x))`.
    Used to represent non-interactive postprocessing.


    Required features: `contrib`

    [make_chain_pm in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_chain_pm.html)

    .. end-markdown

    :param postprocess1: outer postprocessor
    :type postprocess1: Function
    :param measurement0: inner measurement/mechanism
    :type measurement0: Measurement
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_postprocess1 = py_to_c(postprocess1, c_type=Function, type_name=None)
    c_measurement0 = py_to_c(measurement0, c_type=Measurement, type_name=None)

    # Call library function.
    lib_function = lib.opendp_combinators__make_chain_pm
    lib_function.argtypes = [Function, Measurement]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_postprocess1, c_measurement0), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_chain_pm',
            '__module__': 'combinators',
            '__kwargs__': {
                'postprocess1': postprocess1, 'measurement0': measurement0
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_chain_tt(
    transformation1: Transformation,
    transformation0: Transformation
) -> Transformation:
    r"""Construct the functional composition (`transformation1` ○ `transformation0`).
    Returns a Transformation that when invoked, computes `transformation1(transformation0(x))`.


    Required features: `contrib`

    [make_chain_tt in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_chain_tt.html)

    .. end-markdown

    :param transformation1: outer transformation
    :type transformation1: Transformation
    :param transformation0: inner transformation
    :type transformation0: Transformation
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_transformation1 = py_to_c(transformation1, c_type=Transformation, type_name=None)
    c_transformation0 = py_to_c(transformation0, c_type=Transformation, type_name=None)

    # Call library function.
    lib_function = lib.opendp_combinators__make_chain_tt
    lib_function.argtypes = [Transformation, Transformation]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_transformation1, c_transformation0), Transformation))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_chain_tt',
            '__module__': 'combinators',
            '__kwargs__': {
                'transformation1': transformation1, 'transformation0': transformation0
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_composition(
    measurements
) -> Measurement:
    r"""Construct the DP composition \[`measurement0`, `measurement1`, ...\].
    Returns a Measurement that when invoked, computes `[measurement0(x), measurement1(x), ...]`

    All metrics and domains must be equivalent.

    **Composition Properties**

    * sequential: all measurements are applied to the same dataset
    * basic: the composition is the linear sum of the privacy usage of each query
    * noninteractive: all mechanisms specified up-front (but each can be interactive)
    * compositor: all privacy parameters specified up-front (via the map)


    Required features: `contrib`

    .. end-markdown

    :param measurements: A vector of Measurements to compose.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurements = py_to_c(measurements, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["AnyMeasurementPtr"]))

    # Call library function.
    lib_function = lib.opendp_combinators__make_composition
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurements), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurements': measurements
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_fix_delta(
    measurement: Measurement,
    delta: float
) -> Measurement:
    r"""Fix the delta parameter in the privacy map of a `measurement` with a SmoothedMaxDivergence output measure.


    Required features: `contrib`

    [make_fix_delta in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_fix_delta.html)

    .. end-markdown

    :param measurement: a measurement with a privacy curve to be fixed
    :type measurement: Measurement
    :param delta: parameter to fix the privacy curve with
    :type delta: float
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name=None)
    c_delta = py_to_c(delta, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_combinators__make_fix_delta
    lib_function.argtypes = [Measurement, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement, c_delta), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_fix_delta',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement, 'delta': delta
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_fixed_approxDP_to_approxDP(
    measurement: Measurement
) -> Measurement:
    r"""Constructs a new output measurement where the output measure
    is casted from `Approximate<MaxDivergence>` to `SmoothedMaxDivergence`.


    Required features: `contrib`

    [make_fixed_approxDP_to_approxDP in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_fixed_approxDP_to_approxDP.html)

    .. end-markdown

    :param measurement: a measurement with a privacy measure to be casted
    :type measurement: Measurement
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")

    # Call library function.
    lib_function = lib.opendp_combinators__make_fixed_approxDP_to_approxDP
    lib_function.argtypes = [Measurement]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_fixed_approxDP_to_approxDP',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_fully_adaptive_composition(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure
) -> Odometer:
    r"""Construct an odometer that can spawn a compositor queryable.


    Required features: `contrib`

    [make_fully_adaptive_composition in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_fully_adaptive_composition.html)

    **Supporting Elements:**

    * Input Domain    `DI`
    * Input Metric    `MI`
    * Output Measure  `MO`
    * Query           `Measurement<DI, MI, MO, TO>`
    * Answer          `TO`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/combinators/sequential_composition/fully_adaptive/make_fully_adaptive_composition.pdf)

    .. end-markdown

    :param input_domain: indicates the space of valid input datasets
    :type input_domain: Domain
    :param input_metric: how distances are measured between members of the input domain
    :type input_metric: Metric
    :param output_measure: how privacy is measured
    :type output_measure: Measure
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)

    # Call library function.
    lib_function = lib.opendp_combinators__make_fully_adaptive_composition
    lib_function.argtypes = [Domain, Metric, Measure]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure), Odometer))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_fully_adaptive_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_fully_adaptive_composition(
    output_measure: Measure
):  
    r"""partial constructor of make_fully_adaptive_composition

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.combinators.make_fully_adaptive_composition`

    :param output_measure: how privacy is measured
    :type output_measure: Measure
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_fully_adaptive_composition(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure))
    output.__opendp_dict__ = {
            '__function__': 'then_fully_adaptive_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'output_measure': output_measure
            },
        }
    return output



def make_population_amplification(
    measurement: Measurement,
    population_size: int
) -> Measurement:
    r"""Construct an amplified measurement from a `measurement` with privacy amplification by subsampling.
    This measurement does not perform any sampling.
    It is useful when you have a dataset on-hand that is a simple random sample from a larger population.

    The DIA, DO, MI and MO between the input measurement and amplified output measurement all match.


    Required features: `contrib`, `honest-but-curious`

    [make_population_amplification in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_population_amplification.html)

    **Why honest-but-curious?:**

    The privacy guarantees are only valid if the input dataset is a simple sample from a population with `population_size` records.

    .. end-markdown

    :param measurement: the computation to amplify
    :type measurement: Measurement
    :param population_size: the size of the population from which the input dataset is a simple sample
    :type population_size: int
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib", "honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")
    c_population_size = py_to_c(population_size, c_type=ctypes.c_size_t, type_name="usize")

    # Call library function.
    lib_function = lib.opendp_combinators__make_population_amplification
    lib_function.argtypes = [Measurement, ctypes.c_size_t]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement, c_population_size), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_population_amplification',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement, 'population_size': population_size
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_privacy_filter(
    odometer: Odometer,
    d_in,
    d_out
) -> Measurement:
    r"""Combinator that limits the privacy loss of an odometer.

    Adjusts the queryable returned by the odometer
    to reject any query that would increase the total privacy loss
    above the privacy guarantee of the mechanism.


    Required features: `contrib`

    [make_privacy_filter in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_privacy_filter.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `OdometerQueryable<Q, A, MI::Distance, MO::Distance>`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/combinators/privacy_filter/make_privacy_filter.pdf)

    .. end-markdown

    :param odometer: A privacy odometer
    :type odometer: Odometer
    :param d_in: Upper bound on the distance between adjacent datasets
    :param d_out: Upper bound on the privacy loss
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_odometer = py_to_c(odometer, c_type=Odometer, type_name=None)
    c_d_in = py_to_c(d_in, c_type=AnyObjectPtr, type_name=get_distance_type(odometer_input_metric(odometer)))
    c_d_out = py_to_c(d_out, c_type=AnyObjectPtr, type_name=get_distance_type(odometer_output_measure(odometer)))

    # Call library function.
    lib_function = lib.opendp_combinators__make_privacy_filter
    lib_function.argtypes = [Odometer, AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_odometer, c_d_in, c_d_out), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_privacy_filter',
            '__module__': 'combinators',
            '__kwargs__': {
                'odometer': odometer, 'd_in': d_in, 'd_out': d_out
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_pureDP_to_zCDP(
    measurement: Measurement
) -> Measurement:
    r"""Constructs a new output measurement where the output measure
    is casted from `MaxDivergence` to `ZeroConcentratedDivergence`.


    Required features: `contrib`

    [make_pureDP_to_zCDP in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_pureDP_to_zCDP.html)

    **Citations:**

    - [BS16 Concentrated Differential Privacy: Simplifications, Extensions, and Lower Bounds](https://arxiv.org/pdf/1605.02065.pdf#subsection.3.1)

    .. end-markdown

    :param measurement: a measurement with a privacy measure to be casted
    :type measurement: Measurement
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")

    # Call library function.
    lib_function = lib.opendp_combinators__make_pureDP_to_zCDP
    lib_function.argtypes = [Measurement]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_pureDP_to_zCDP',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_select_private_candidate(
    measurement: Measurement,
    stop_probability: float,
    threshold: float
) -> Measurement:
    r"""Select a private candidate whose score is above a threshold.

    Given `measurement` that satisfies ε-DP, returns new measurement M' that satisfies 2ε-DP.
    M' releases the first invocation of `measurement` whose score is above `threshold`.

    Each time a score is below `threshold`
    the algorithm may terminate with probability `stop_probability` and return nothing.

    `measurement` should make releases in the form of (score, candidate).
    If you are writing a custom scorer measurement in Python,
    specify the output type as `TO=(float, "ExtrinsicObject")`.
    This ensures that the float value is accessible to the algorithm.
    The candidate, left as arbitrary Python data, is held behind the ExtrinsicObject.

    Algorithm 1 in `Private selection from private candidates <https://arxiv.org/pdf/1811.07971.pdf#page=7>`_ (Liu and Talwar, STOC 2019).


    Required features: `contrib`

    [make_select_private_candidate in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_select_private_candidate.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `Option<(f64, TO)>`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/combinators/select_private_candidate/make_select_private_candidate.pdf)

    .. end-markdown

    :param measurement: A measurement that releases a 2-tuple of (score, candidate)
    :type measurement: Measurement
    :param stop_probability: The probability of stopping early at any iteration.
    :type stop_probability: float
    :param threshold: The threshold score. Return immediately if the score is above this threshold.
    :type threshold: float
    :return: A measurement that returns a release from `measurement` whose score is greater than `threshold`, or none.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")
    >>> threshold = 23
    >>> space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    ...
    >>> # For demonstration purposes-- construct a measurement that releases
    >>> # a tuple with a differentially private score and value.
    >>> # The tuple released must satisfy the privacy guarantee from the map.
    >>> import numpy as np
    >>> m_mock = space >> dp.m.then_user_measurement(
    ...     dp.max_divergence(),
    ...     lambda x: (np.random.laplace(loc=x), "arbitrary candidate"),
    ...     lambda d_in: d_in,
    ...     TO="(f64, ExtrinsicObject)"
    ... )
    ...
    >>> m_private_selection = dp.c.make_select_private_candidate(
    ...     m_mock, threshold=threshold, stop_probability=0
    ... )
    ...
    >>> score, candidate = m_private_selection(20)
    ...
    >>> assert score >= threshold
    >>> assert m_private_selection.map(1) == 2 * m_mock.map(1)
    >>> assert isinstance(candidate, str)


    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")
    c_stop_probability = py_to_c(stop_probability, c_type=ctypes.c_double, type_name="f64")
    c_threshold = py_to_c(threshold, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_combinators__make_select_private_candidate
    lib_function.argtypes = [Measurement, ctypes.c_double, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement, c_stop_probability, c_threshold), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_select_private_candidate',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement, 'stop_probability': stop_probability, 'threshold': threshold
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


@deprecated(version="0.14.0", reason="This function has been renamed, use `make_adaptive_composition` instead.")
def make_sequential_composition(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    d_in,
    d_mids
) -> Measurement:
    r"""Construct a Measurement that when invoked,
    returns a queryable that interactively composes measurements.

    **Composition Properties**

    * sequential: all measurements are applied to the same dataset
    * basic: the composition is the linear sum of the privacy usage of each query
    * interactive: mechanisms can be specified based on answers to previous queries
    * compositor: all privacy parameters specified up-front

    If the privacy measure supports concurrency,
    this compositor allows you to spawn multiple interactive mechanisms
    and interleave your queries amongst them.


    Required features: `contrib`

    [make_sequential_composition in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_sequential_composition.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `Queryable<Measurement<DI, MI, MO, TO>, TO>`

    .. end-markdown

    :param input_domain: indicates the space of valid input datasets
    :type input_domain: Domain
    :param input_metric: how distances are measured between members of the input domain
    :type input_metric: Metric
    :param output_measure: how privacy is measured
    :type output_measure: Measure
    :param d_in: maximum distance between adjacent input datasets
    :param d_mids: maximum privacy expenditure of each query
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    QO = get_distance_type(output_measure) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_d_in = py_to_c(d_in, c_type=AnyObjectPtr, type_name=get_distance_type(input_metric))
    c_d_mids = py_to_c(d_mids, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=[QO]))

    # Call library function.
    lib_function = lib.opendp_combinators__make_sequential_composition
    lib_function.argtypes = [Domain, Metric, Measure, AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_d_in, c_d_mids), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_sequential_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'd_in': d_in, 'd_mids': d_mids
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_sequential_composition(
    output_measure: Measure,
    d_in,
    d_mids
):  
    r"""partial constructor of make_sequential_composition

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.combinators.make_sequential_composition`

    :param output_measure: how privacy is measured
    :type output_measure: Measure
    :param d_in: maximum distance between adjacent input datasets
    :param d_mids: maximum privacy expenditure of each query
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_sequential_composition(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        d_in=d_in,
        d_mids=d_mids))
    output.__opendp_dict__ = {
            '__function__': 'then_sequential_composition',
            '__module__': 'combinators',
            '__kwargs__': {
                'output_measure': output_measure, 'd_in': d_in, 'd_mids': d_mids
            },
        }
    return output



def make_zCDP_to_approxDP(
    measurement: Measurement
) -> Measurement:
    r"""Constructs a new output measurement where the output measure
    is casted from `ZeroConcentratedDivergence` to `SmoothedMaxDivergence`.


    Required features: `contrib`

    [make_zCDP_to_approxDP in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/combinators/fn.make_zCDP_to_approxDP.html)

    .. end-markdown

    :param measurement: a measurement with a privacy measure to be casted
    :type measurement: Measurement
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_measurement = py_to_c(measurement, c_type=Measurement, type_name="AnyMeasurement")

    # Call library function.
    lib_function = lib.opendp_combinators__make_zCDP_to_approxDP
    lib_function.argtypes = [Measurement]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_measurement), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_zCDP_to_approxDP',
            '__module__': 'combinators',
            '__kwargs__': {
                'measurement': measurement
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output



# type: ignore
# MaxDivergence
def noisy_top_k(x: list[TIA], scale: f64, k: usize, negate: bool) -> list[usize]:
    return exponential_top_k(x, scale, k, negate)

def privacy_map(d_in: f64, scale: f64) -> f64:
    return d_in.inf_div(scale)



# type: ignore
def find_nearest_multiple_of_2k(x: RBig, k: i32) -> IBig:
    # exactly compute x/2^k and break into fractional parts
    num, den = x_mul_2k(x, -k).into_parts() # `\label{line:into-parts}`

    # argmin_i |i * 2^k - x|, the index of nearest multiple of 2^k
    return (floor_div(num << 1, den) + 1) >> 1 # `\label{line:return}`



# type: ignore
class FloatExpFamily:
    def make_noise_threshold(
        self,
        input_space: tuple[
            MapDomain[AtomDomain[TK], MapDomain[TV]], L0PInfDistance[P, AbsoluteDistance[QI]]
        ],
        threshold: TV,
    ) -> Measurement[
        MapDomain[AtomDomain[TK], AtomDomain[TV]],
        HashMap[TK, TV],
        L0PInfDistance[P, AbsoluteDistance[QI]],
        MO,
    ]:
        scale, k = self.scale, self.k
        distribution = ZExpFamily(
            scale=integerize_scale(scale, k)
        )  # |\label{line:dist}|

        if threshold.is_sign_negative():
            raise f"threshold ({threshold}) must not be negative"

        r_threshold = RBig.try_from(threshold)
        r_threshold = x_mul_2k(r_threshold, -k).round()

        t_int = make_float_to_bigint_threshold(input_space, threshold, k)
        m_noise = distribution.make_noise_threshold(t_int.output_space(), r_threshold)
        return t_int >> m_noise >> then_deintegerize_hashmap(k)



# type: ignore
def truncate_domain(
    domain: DslPlanDomain,
    truncation: Truncation,
) -> DslPlanDomain:
    match truncation:
        case Truncation.Filter(_):
            for m in domain.margins:
                # After filtering you no longer know partition lengths or keys.
                m.invariant = None  # |\label{line:invariant-filter}|
            # Upper bounds on the number of rows and groups remain valid.
            return domain
        case Truncation.GroupBy(keys, aggs):
            for agg in aggs:
                # True means resize is allowed
                check_infallible(agg, True)  # |\label{line:infallible}|

            def with_truncation(lf):
                return lf.group_by(keys).agg(aggs)

            def without_invariant(m):
                m.invariant = None
                return m

            # derive new output domain
            return FrameDomain.new_with_margins(
                [  # |\label{line:simulate-schema}|
                    Seriesdomain.new_from_field(f)
                    for f in domain.simulate_schema(with_truncation)
                ],
                margins=[
                    # discard invariants as multiverses are mixed
                    without_invariant(m.clone())
                    for m in domain.margins
                    # only keep margins that are a subset of the truncation keys
                    if m.by.is_subset(HashSet.from_iter(keys))
                ],
            )



# type: ignore
def match_num_groups_predicate(
    ranks: Expr,
    partition_by: Vec[Expr],
    identifier: Expr,
    threshold: u32,
) -> Optional[Bound]:
    # check if is a rank function
    if not isinstance(ranks, Expr.Function) or not isinstance(
        ranks.function, FunctionExpr.Rank
    ):  # |\label{line:check_rank}|
        return None

    input = ranks.input
    options = ranks.function.options

    if partition_by != [identifier]:  # |\label{line:check_identifier}|
        raise "num_groups truncation must use the identifier in the over clause"

    if not isinstance(options.method, RankMethod.Dense):  # |\label{line:check_method}|
        raise "num_groups truncation's rank must be dense"

    if len(input) != 1:  # |\label{line:check_input}|
        raise "rank function must be applied to a single input"

    input_item = input[0]

    # Treat as_struct as a special case that represents multiple columns.
    if isinstance(input_item, Expr.Function) and isinstance(
        input_item.function, FunctionExpr.AsStruct
    ):  # |\label{line:extract_grouping_columns}|

        # If the first field is a hash of the second field,
        # then interpret the grouping columns as the hash input.
        # The second field disambiguates hash collisions when ranking.
        if isinstance(input_item, Expr.Function) and isinstance(
            input_item.function, FunctionExpr.Hash
        ):  # |\label{line:check_hash}|
            hash_input = input_item.input
            if hash_input.get(0) == input.get(1):
                if not isinstance(hash_input.get(0), Expr.Function) or not isinstance(
                    hash_input.get(0).function, FunctionExpr.AsStruct
                ):
                    raise f"expected hash input to be a struct, found {hash_input}"
                input = hash_input.get(0)

        by = set(input_item.input)
    else:
        by = {input_item}

    return Bound(by=by, per_group=None, num_groups=threshold)  # |\label{line:bound}|



# type: ignore
# analogous to impl NoiseThresholdPrivacyMap<L01InfDistance<AbsoluteDistance<RBig>>, Approximate<MaxDivergence>> for ZExpFamily<1> in Rust
class ZExpFamily1:
    def noise_threshold_privacy_map(
        self,
        _input_metric: L01InfDistance[AbsoluteDistance[RBig]],
        output_measure: Approximate[MaxDivergence],
        threshold: UBig,
    ) -> PrivacyMap[L01InfDistance[AbsoluteDistance[RBig]], Approximate[MaxDivergence]]:
        # |\label{line:noise-privacy-map}|
        noise_privacy_map = self.noise_privacy_map(L1Distance.default(), output_measure[0])
        scale = self.scale

        def privacy_map(d_in: tuple[u32, RBig, RBig]):
            l0, l1, li = d_in

            l1_sign, l1 = l1.floor().into_parts() # |\label{line:l1-floor}|
            if l1_sign != Sign.Positive: # |\label{line:l1-check}|
                raise f"l1 sensitivity ({l1}) must be non-negative"

            li_sign, li = li.floor().into_parts() # |\label{line:li-floor}|
            if li_sign != Sign.Positive: # |\label{line:li-check}|
                raise f"l-infinity sensitivity ({li}) must be non-negative"
            
            l1 = l1.min(li * l0)  # |\label{line:l1}|
            li = li.min(l1)  # |\label{line:li}|

            if l1.is_zero():  # |\label{line:zero-sens}|
                return 0.0, 0.0

            if scale.is_zero(): # |\label{line:zero-scale}|
                return f64.INFINITY, 1.0

            epsilon = noise_privacy_map.eval(l1) # |\label{line:epsilon}|
            
            if li > threshold: # |\label{line:threshold-check}|
                raise f"threshold must not be smaller than {li}"
            
            d_instability = threshold - li # |\label{line:distance-to-instability}|

            try:
                alpha_disc = conservative_discrete_laplacian_tail_to_alpha(
                    scale,
                    d_instability
                )
            except Exception:
                alpha_disc = None
            
            try:
                alpha_cont = conservative_continuous_laplacian_tail_to_alpha(
                    scale,
                    d_instability,
                )
            except Exception:
                alpha_cont = None

            delta_single = option_min(alpha_disc, alpha_cont)
            if delta_single is None:
                raise "failed to compute tail bound in privacy map"

            delta_joint: f64 = (1.0).inf_sub(
                (1.0).neg_inf_sub(delta_single).neg_inf_powi(IBig.from_(l0)),
            )

            # delta is only sensibly at most 1
            return epsilon, delta_joint.min(1.0)

        return PrivacyMap.new_fallible(privacy_map)



# type: ignore
def sample_discrete_gaussian(scale) -> int:
    if scale == 0:
        return 0
    
    t = floor(scale) + 1 # |\label{line:t}|
    sigma2 = scale**2
    
    while True:
        candidate = sample_discrete_laplace(t) # |\label{line:candidate}|
        x = abs(candidate) - sigma2 / t
        bias = x**2 / (2 * sigma2)  # |\label{line:bias}|
        if sample_bernoulli_exp(bias): # |\label{line:bern}|
            return candidate


import pytest
import opendp.prelude as dp


def test_binary_search_fail():
    with pytest.raises(ZeroDivisionError) as e:
        dp.binary_search(lambda _: bool(1 / 0), T=float)
    if hasattr(e.value, "add_note"):
        assert e.value.__notes__[0] == "Predicate in binary search always raises an exception. This exception is raised when the predicate is evaluated at 0.0." # type: ignore[attr-defined]

def test_binary_search_overflow():

    input_domain = dp.vector_domain(dp.atom_domain(T=float, nan=False))
    input_metric = dp.symmetric_distance()

    d_in = 1
    d_out = 1.01
    bounded_sum = (
        dp.t.make_clamp(input_domain, input_metric, bounds=(0.0, 1.0)) >>
        dp.t.then_sum()
    )
    dp.binary_search_param(
        lambda s: bounded_sum >> dp.m.then_laplace(scale=s),
        d_in=d_in,
        d_out=d_out
    )

def test_stuck():
    input_space = dp.vector_domain(dp.atom_domain(T=float, nan=False)), dp.symmetric_distance()

    epsilon = 1.3
    sens = 500_000.0 * 500_000.0
    bounded_sum = (
        input_space >>
        dp.t.then_clamp(bounds=(0.0, sens)) >>
        dp.t.then_sum()
    )
    real_v = sens / epsilon
    discovered_scale = dp.binary_search_param(
        lambda s: bounded_sum >> dp.m.then_laplace(scale=s),
        d_in=1,
        bounds=(0.0, real_v * 2.0),
        d_out=epsilon)
    assert discovered_scale == 192307694098.69745
    
def test_binary_search():
    assert dp.binary_search(lambda x: x <= -5, T=int) == -5
    assert dp.binary_search(lambda x: x <= 5, T=int) == 5
    assert dp.binary_search(lambda x: x >= -5, T=int) == -5
    assert dp.binary_search(lambda x: x >= 5, T=int) == 5


def test_binary_search_inferred_int():
    def predicate(v):
        if not isinstance(v, int):
            raise TypeError("Expected int")
        return v <= 5
    
    assert dp.binary_search(predicate) == 5

def test_binary_search_inferred_int_fail():
    def predicate(v):
        if not isinstance(v, int):
            raise TypeError("Expected int")
        raise ValueError("I always fail for some other reason")
    
    # int is inferred, but the ValueError should pop up
    with pytest.raises(ValueError):
        dp.binary_search(predicate)

def test_binary_search_failed_infer():
    def predicate(_):
        raise TypeError("always raised for some other reason")
    
    # type inference should fail, and we get the error from type inference
    with pytest.raises(TypeError):
        dp.binary_search(predicate)

    
def test_type_inference():
    def chainer(b):
        return dp.t.make_sum(
            dp.vector_domain(dp.atom_domain(bounds=(-b, b)), size=1000), 
            dp.symmetric_distance())
    assert dp.binary_search_param(chainer, 2, 100) == pytest.approx(50)

    def mean_chainer_n(n):
        return dp.t.make_mean(
            dp.vector_domain(dp.atom_domain(bounds=(-20., 20.)), size=n), 
            dp.symmetric_distance())
    assert dp.binary_search_param(mean_chainer_n, 2, 1.) == 41

    def mean_chainer_b(b):
        return dp.t.make_mean(
            dp.vector_domain(dp.atom_domain(bounds=(-b, b)), size=1000), 
            dp.symmetric_distance())
    assert 499.999 < dp.binary_search_param(mean_chainer_b, 2, 1.) < 500.




# type: ignore
def sample_uniform_ubig_below(upper: UBig) -> UBig:
    byte_len = upper.bit_len().div_ceil(8)
    max = Ubig.from_be_bytes([u8.MAX] * byte_len)
    threshold = max - max % upper

    buffer = [0] * byte_len
    while True:
        fill_bytes(buffer)

        sample = UBig.from_be_bytes(buffer)
        if sample < threshold:
            return sample % upper



"""High-level mechanism for applying mbi mechanisms to dataframes with mixed types."""

from math import sqrt
from dataclasses import dataclass, field
from typing import (
    Any,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypeAlias,
    Union,
    cast,
    TYPE_CHECKING,
)
from opendp._internal import _new_pure_function
from opendp._lib import import_optional_dependency
from opendp.combinators import (
    make_adaptive_composition,
    make_approximate,
    make_composition,
)
from opendp.extras.mbi._aim import AIM
from opendp.extras.mbi._utilities import (
    prior,
    get_std,
    row_major_order,
    weight_marginals,
    Algorithm,
    ONEWAY_UNKEYED,
)
from opendp.measurements import make_private_lazyframe
from opendp.mod import (
    ApproximateDivergence,
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Measurement,
    OpenDPException,
    Queryable,
    binary_search,
    binary_search_param,
)
from opendp.domains import _lazyframe_from_domain
from opendp.transformations import make_stable_lazyframe
from opendp.typing import RuntimeType


if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound


@dataclass
class ContingencyTable:
    """Approximate a k-way contingency table via low-order marginals."""

    keys: dict[str, Any]  # dict[str, Series]
    """Unique keys for each column"""
    cuts: dict[str, Any]  # dict[str, Series]
    """Cut points (left-open) for each numeric column"""
    marginals: dict[tuple[str, ...], Any]  # dict[tuple[str, ...], LinearMeasurement]
    """Mapping from clique name to LinearMeasurement used to fit the model"""
    model: Any  # MarkovRandomField
    """MarkovRandomField spanning the same columns as keys"""
    thresholds: dict[str, int] = field(default_factory=dict)
    """Cut-off point for discovered stable keys. 
    Any category appearing fewer than threshold times is attributed to the null category."""

    def __post_init__(self):
        from mbi import MarkovRandomField, Domain  # type: ignore[import-untyped,import-not-found]
        import polars as pl  # type: ignore[import-not-found]

        if not isinstance(self.model, MarkovRandomField):  # pragma: no cover
            raise ValueError("model must be a MarkovRandomField")

        for col in self.keys:
            self.keys[col] = pl.Series(col, self.keys[col])
        for col in self.cuts:
            self.cuts[col] = pl.Series(col, self.cuts[col])

        key_shape = set((c, len(k)) for c, k in self.keys.items())
        if set(self.model.domain.config.items()) != key_shape:
            raise ValueError(
                "Model domain must match key attrs and sizes.\n"
                f"  Model shape: {self.model.domain}\n"
                f"  Key shape:   {Domain.fromdict(dict(key_shape))}"
            )

        for col, cutset in self.cuts.items():
            if col not in self.keys:
                raise ValueError(f'"{col}" in cuts is not present in keys')
            if len(self.keys[col]) - 1 != len(cutset):
                msg = f'"{col}" keyset length ({len(self.keys[col])}) must be one greater than the number of cuts ({len(cutset)})'
                raise ValueError(msg)

    def synthesize(
        self, rows: Optional[int] = None, method: Literal["round", "sample"] = "round"
    ):
        """Generate synthetic data that is consistent with the contingency table.

        :param rows: if not set, approximately matches number of rows in original data
        :param method: "round" rounds to projection, "sample" samples from densities
        """
        import polars as pl  # type: ignore[import-not-found]
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        model = cast(MarkovRandomField, self.model)

        indices = model.synthetic_data(rows, method).df
        data = {
            c: _deindex(indices[c].to_numpy(), self.keys[c], cuts=self.cuts.get(c))
            for c in self.keys
        }
        return pl.DataFrame(data)

    def project(self, attrs: str | Sequence[str]):
        """Return counts corresponding to each combination of self.keys in `attrs` in an ndarray.

        :param attrs: attributes to preserve. All other attributes are marginalized.
        """
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        model = cast(MarkovRandomField, self.model)
        return model.project(attrs).values

    def project_melted(self, attrs: str | Sequence[str]):
        """Return counts corresponding to each combination of self.keys in `attrs` in a dataframe.

        :param attrs: attributes to preserve. All other attributes are marginalized.
        """
        import polars as pl
        import numpy as np

        attrs = [attrs] if isinstance(attrs, str) else list(attrs)

        lengths_arr = np.asarray(self.project(attrs).ravel())
        lengths = pl.Series("len", lengths_arr).to_frame()

        return row_major_order(self.keys[attr] for attr in attrs).hstack(lengths)

    def std(self, attrs: Sequence[str]) -> float:
        """Estimate the standard deviation of the count estimate.
        The estimate does not take cuts into account.

        When marginals are estimated with gaussian noise,
        the consistent (variance-weighted) marginals are also gaussian-distributed.
        Therefore the standard deviation matches the noise scale parameter.

        When marginals are estimated with laplace noise,
        the consistent marginals are a sum of laplacians, which is not laplace-distributed.
        Therefore the standard deviation doesn't correspond to the laplace distribution.
        Under the central limit theorem, the distribution will tend towards gaussian.

        If the estimate is gaussian-distributed,
        use :py:func:`opendp.accuracy.gaussian_scale_to_accuracy`
        to construct a confidence interval.

        :param attrs: attributes to preserve in uncertainty estimate
        """
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        model = cast(MarkovRandomField, self.model)

        if isinstance(attrs, str):  # pragma: no cover
            raise ValueError(f"attrs ({attrs}) must be a sequence of strings")

        attrs_clique = set(attrs)
        size = model.domain.size

        inv_var_sum = sum(
            1 / (lm.stddev**2 * size(clique) / size(attrs_clique))
            for clique, lm in self.marginals.items()
            if attrs_clique.issuperset(clique)
        )

        if inv_var_sum == 0:
            raise ValueError(f"attrs ({attrs}) are not covered by the query set")

        return sqrt(1 / inv_var_sum)

    @property
    def schema(self) -> dict[str, Any]:
        """Returns the data schema."""
        return {keyset.name: keyset.dtype for keyset in self.keys.values()}


def make_contingency_table(
    input_domain: LazyFrameDomain,
    input_metric: FrameDistance,
    output_measure: Measure,
    d_in: list["Bound"],
    d_out: Union[float, tuple[float, float]],
    keys: Optional[Mapping[str, Sequence]] = None,
    cuts: Optional[Mapping[str, Sequence[float]]] = None,
    table: Optional[ContingencyTable] = None,
    algorithm: Algorithm = AIM(),
) -> tuple[Measurement, Optional[float], Optional[int]]:
    """Return a measurement that releases a :class:`.ContingencyTable`,
    as well as the noise scale and optional threshold for one-way marginals.

    :param input_domain: domain of input data
    :param input_metric: how to compute distance between datasets
    :param output_measure: how to measure privacy of release
    :param d_in: bounds on distances between adjacent input datasets
    :param d_out: upper bound on the privacy loss
    :param keys: dictionary of column names and unique categories
    :param cuts: dictionary of column names and left-open bin edges
    :param table: ContingencyTable from prior release
    :param algorithm: configuration for internal estimation algorithm
    """
    import_optional_dependency("mbi")
    import polars as pl
    import mbi

    plan = _lazyframe_from_domain(input_domain)
    schema = plan.collect_schema()

    keys_pl: dict[str, pl.Series] = {
        col: _unique(_with_null(pl.Series(col, keyset, dtype=schema.get(col))), "keys")
        for col, keyset in (keys or dict()).items()
    }

    cuts_pl: dict[str, pl.Series] = {
        col: _unique(_increasing(pl.Series(col, cutset, dtype=schema.get(col))), "cuts")
        for col, cutset in (cuts or dict()).items()
    }

    # add cut bin labels to keys
    def get_categories(cutset):
        labels = [f"({lb}, {rb}]" for lb, rb in zip(["-inf", *cutset], [*cutset, "inf"])]
        return pl.Series(cutset.name, labels)

    keys_pl |= {col: get_categories(cutset) for col, cutset in cuts_pl.items()}

    if table:
        keys_pl |= table.keys
        cuts_pl |= table.cuts
        model = cast(mbi.MarkovRandomField, table.model)
        marginals = table.marginals.copy()
        thresholds = table.thresholds.copy()
    else:
        model = None
        marginals = {}
        thresholds = {}

    if cuts_pl:
        plan = plan.with_columns(pl.col(c).cut(cutset, labels=get_categories(cutset)) for c, cutset in cuts_pl.items() if c in schema)  # type: ignore[arg-type]

    if (QO := RuntimeType.infer(d_out)) != output_measure.distance_type:
        raise ValueError(f"d_out type ({QO}) must be {output_measure.distance_type}")

    delta = None
    if isinstance(d_out, tuple):
        d_out, delta = d_out

    all_keys_known = set(input_domain.columns) <= set(keys_pl)
    if delta and all_keys_known:
        message = f"delta ({delta}) must be zero because keys and cuts span all columns"
        raise ValueError(message)
    if not delta and not all_keys_known:
        message = f"delta ({delta}) must be nonzero because keys and cuts don't span all columns"
        raise ValueError(message)

    if algorithm.oneway == ONEWAY_UNKEYED and all_keys_known:
        d_multiway: Union[float | tuple[float, float]] = d_out
        d_mids = [d_multiway if delta is None else (d_multiway, 0.0)]
        m_oneway = None
        scale, threshold = None, None

    else:
        if algorithm.oneway_split is None:
            oneway_split = 0.5
            if algorithm.oneway == ONEWAY_UNKEYED:
                columns = set(input_domain.columns)
                oneway_split *= len(columns - set(keys_pl)) / len(columns)
        else:
            oneway_split = algorithm.oneway_split

        d_oneway = d_out * oneway_split
        d_multiway = float(prior(prior(d_out - d_oneway)))
        if delta is not None:
            d_oneway = d_oneway, delta  # type: ignore[assignment]

        d_mids = [d_oneway, d_multiway if delta is None else (d_multiway, 0.0)]  # type: ignore[has-type]
        m_oneway, scale, threshold = _make_oneway_marginals(
            input_domain,
            input_metric,
            output_measure,
            d_in=d_in,
            d_out=d_oneway,
            keys=keys_pl,
            plan=plan,
            unknown_only=algorithm.oneway == ONEWAY_UNKEYED,
        )
        if threshold is not None:
            thresholds |= {
                c: threshold for c in input_domain.columns if c not in keys_pl
            }

    m_comp = make_adaptive_composition(
        input_domain, input_metric, output_measure, d_in, d_mids
    )

    if isinstance(output_measure, ApproximateDivergence):
        inner_measure = output_measure.inner_measure
    else:
        inner_measure = output_measure

    if "len" in input_domain.columns:
        raise ValueError('input_domain must not contain a column named "len"')

    def postprocess(qbl: Queryable) -> ContingencyTable:
        stable_keys = keys_pl.copy()
        current_marginals = marginals.copy()

        if m_oneway:
            new_keys, oneway_releases = qbl(m_oneway)
            stable_keys |= new_keys
            current_marginals = weight_marginals(current_marginals, *oneway_releases)

        mbi_domain = mbi.Domain.fromdict({c: len(k) for c, k in stable_keys.items()})

        potentials = None
        if isinstance(model, mbi.MarkovRandomField):
            import attr  # type: ignore[import-not-found]

            potential_factor = attr.evolve(model.potentials, domain=mbi_domain)
            potentials = potential_factor.expand(list(current_marginals.keys()))

        current_model = algorithm.estimator(
            mbi_domain, list(current_marginals.values()), potentials=potentials
        )

        t_index = make_stable_lazyframe(
            input_domain,
            input_metric,
            plan.with_columns(
                pl.col(c)
                .replace_strict(
                    stable_keys[c],
                    pl.Series(range(len(stable_keys[c])), dtype=pl.UInt32),
                    default=_get_null_index(stable_keys[c], len(stable_keys[c]) - 1),
                )
                .clip(0, len(stable_keys[c]) - 1)
                for c in input_domain.columns
            ),
        )

        m_marginals = algorithm.make_marginals(
            input_domain=t_index.output_domain,
            input_metric=t_index.output_metric,
            output_measure=inner_measure,
            d_in=t_index.map(d_in),
            d_out=d_multiway,  # type: ignore[arg-type]
            marginals=current_marginals,
            model=current_model,
        )

        m_index_marginals = t_index >> m_marginals

        if delta is not None:
            m_index_marginals = make_approximate(m_index_marginals)

        T: TypeAlias = tuple[dict[tuple[str, ...], Any], mbi.MarkovRandomField]
        current_marginals, current_model = cast(T, qbl(m_index_marginals))

        ordered_keys = {c: stable_keys[c] for c in input_domain.columns}
        ordered_keys |= stable_keys

        return ContingencyTable(
            keys=ordered_keys,
            cuts=cuts_pl,
            marginals=current_marginals,
            model=current_model,
            thresholds=thresholds,
        )

    return m_comp >> _new_pure_function(postprocess), scale, threshold


def _unique(series, arg: str):
    """raise error if elements are not unique"""
    if series.n_unique() != series.len():  # pragma: no cover
        # this is actually tested, but coverage doesn't see it
        raise ValueError(f'{arg} must be unique: "{series.name}" has duplicates')
    return series


def _increasing(series):
    """raise error if elements are not increasing"""
    import polars as pl

    diff = series.cast(pl.Float64).diff(null_behavior="drop")
    if diff.le(0).any():  # pragma: no cover
        # this is actually tested, but coverage doesn't see it
        message = f'cuts must be strictly increasing: "{series.name}" is not strictly increasing'
        raise ValueError(message)
    return series


def _with_null(series):
    """create new series leading with null if null is not present"""
    return (
        series if series.is_null().any() else series.clone().append(series.clear(n=1))
    )


def _get_null_index(series, default=None) -> Optional[int]:
    """returns the index of null in a polars `series`, else `default`"""
    null_indices = series.is_null().arg_true()
    return null_indices.item() if null_indices.len() else default


def _make_oneway_marginals(
    input_domain: LazyFrameDomain,
    input_metric: FrameDistance,
    output_measure: Measure,
    d_in,
    d_out,
    keys: dict[str, Any],
    plan: Any,
    unknown_only: bool,
) -> tuple[Measurement, float, Optional[int]]:
    """Returns a measurement that releases keys and counts for one-way marginals,
    as well as the discovered scale and threshold."""
    from opendp.extras.polars import dp_len
    import polars as pl  # type: ignore[import-not-found]
    from mbi import LinearMeasurement  # type: ignore[import-not-found]
    from mbi.estimation import minimum_variance_unbiased_total  # type: ignore[import-untyped,import-not-found]
    import numpy as np  # type: ignore[import-not-found]

    def group_by_agg(plan: pl.LazyFrame, name: str) -> pl.LazyFrame:
        """Aggregate the compute plan"""
        keyset = keys.get(name)
        if keyset is None:
            return plan.group_by(name).agg(dp_len())

        # fold all unknown values into null
        if isinstance(keyset.dtype, pl.Categorical):  # pragma: no cover
            keyset = keyset.cast(pl.Enum(keyset.cast(pl.String)))
        replace = pl.col(name).replace_strict(keys[name], keyset, default=None)

        return pl.LazyFrame([keyset]).join(
            plan.select(replace).group_by(name).agg(dp_len()),
            on=[name],
            how="left",
            nulls_equal=True,
        )

    def _make(scale: float, threshold: Optional[int] = None) -> Measurement:
        std = get_std(output_measure, scale)

        # all columns that should be estimated
        names = [
            name
            for name in input_domain.columns
            if name not in keys or not unknown_only
        ]
        # estimating the total is necessary if there are no keyed columns
        should_total = all(name not in keys for name in input_domain.columns)

        one_way_measurements = [
            make_private_lazyframe(
                input_domain,
                input_metric,
                output_measure,
                group_by_agg(plan, name),
                global_scale=scale,
                threshold=threshold,
            )
            >> _new_pure_function(lambda x: x.collect())
            for name in names
        ]
        if should_total:
            m_total = make_private_lazyframe(
                input_domain,
                input_metric,
                output_measure,
                plan.select(dp_len()),
                global_scale=scale,
            ) >> _new_pure_function(lambda x: x.collect())
            one_way_measurements.append(m_total)

        def postprocess(counts: list[pl.DataFrame]):
            new_keys = dict()
            lm_counts = []

            if should_total:
                total = counts.pop().item()
            else:
                total = minimum_variance_unbiased_total(
                    LinearMeasurement(count["len"].to_numpy(), [name], stddev=std)
                    for name, count in zip(names, counts)
                    if name in keys
                )

            for name, count in zip(names, counts):
                count = count.sort(name)
                lookup = dict(zip(count[name], count["len"]))

                if name in keys:
                    values_iter: Iterator[int] = (lookup[k] for k in keys[name])
                else:
                    lookup.setdefault(None, 0)
                    lookup[None] += max(total - count["len"].sum(), 0)
                    dtype = count[name].dtype

                    new_keys[name] = pl.Series(name, lookup.keys(), dtype=dtype)
                    values_iter = lookup.values()  # type: ignore[assignment]

                values = np.fromiter(values_iter, dtype=np.uint32)
                lm_counts.append(LinearMeasurement(values, clique=[name], stddev=std))

            return new_keys, lm_counts

        return make_composition(one_way_measurements) >> _new_pure_function(postprocess)

    if isinstance(output_measure, ApproximateDivergence):
        compare_s = lambda s: _make(s, threshold=2**32 - 1).map(d_in)[0] < d_out[0]  # type: ignore[index]
        scale = binary_search(compare_s, T=float)

        try:
            _make(scale, threshold=None)
            threshold: Optional[int] = None  # pragma: no cover
        except OpenDPException:
            compare_t = lambda t: _make(scale, t).map(d_in)[1] < d_out[1]  # type: ignore[index,arg-type]
            threshold = binary_search(compare_t, T=int)  # type: ignore[assignment]

    else:
        scale = binary_search_param(_make, d_in, d_out, T=float)
        threshold = None

    return _make(scale, threshold), scale, threshold


def _deindex(indices, keys, cuts=None):
    """Translate keyset indices into categorical data."""
    import numpy as np

    if cuts is None:
        return keys[indices]

    if cuts.dtype.is_integer():
        # cuts are left-open
        sampler = lambda lower, upper: np.random.randint(lower + 1, upper + 1)
    elif cuts.dtype.is_float():
        sampler = np.random.uniform  # type: ignore[assignment]

    edges = np.array([cuts[0] - 1, *cuts, cuts[-1] + 1])
    return sampler(edges[indices], edges[indices + 1])



# type: ignore
def make_clamp(
    input_domain: VectorDomain[AtomDomain[TA]], 
    input_metric: M, 
    bounds: tuple[TA, TA]
): # |\label{line:def}|
    input_domain.element_domain.assert_non_null() # |\label{line:assert-non-null}|

    # clone to make it explicit that we are not mutating the input domain
    output_row_domain = input_domain.element_domain.clone()
    output_row_domain.bounds = Bounds.new_closed(bounds)

    def clamper(value: TA) -> TA: # |\label{line:clamper}|
        return value.total_clamp(bounds[0], bounds[1])
    
    return make_row_by_row_fallible( # |\label{line:row-by-row}|
        input_domain, 
        input_metric, 
        output_row_domain, 
        clamper
    )


# type: ignore
def sample_bernoulli_exp1(x) -> bool:
    k = 1
    while True:
        if sample_bernoulli_rational(x / k, false): # |\label{line:B_i}|
            k += 1
        else: 
            return is_odd(k) # |\label{line:K}|



# type: ignore
def new_continuation_rule(d_in: MI_Distance, d_out: MO_Distance) -> Wrapper:

    def wrapper(queryable: Queryable) -> Queryable:
        state = queryable
        del queryable  # (moved by rust ownership)

        def transition(query: Query[Any]) -> Answer[Any]:  # `\label{transition}`

            if state is None:
                raise "filter is exhausted"
            else:
                queryable = state

            answer = queryable.eval_query(query)

            match queryable.eval_poly(OdometerQuery.PrivacyLoss(d_in.clone())):  # `\label{external}`
                case OdometerAnswer.PrivacyLoss(pending_d_out):
                    pass
                case _:
                    raise "expected privacy loss"

            if pending_d_out.total_gt(d_out):  # `\label{check}`
                state.take()  # `\label{free}`
                raise "filter is now exhausted"

            return answer   # `\label{return}`
            
        return Queryable.new_raw(transition)

    return Wrapper.new(wrapper)  # `\label{wrapper}`



import re
from typing import List, Tuple, Any

import pytest

from opendp.mod import *
from opendp.typing import *
from opendp.typing import _INTEGER_TYPES


def test_numpy_function():
    np = pytest.importorskip('numpy')
    assert str(RuntimeType.infer(np.array([1, 2, 3]))) == 'Vec<i64>'
    assert str(RuntimeType.infer(np.array(1))) == 'i32'
    assert str(RuntimeType.infer(np.array(1.))) == 'f64'
    assert str(RuntimeType.infer(np.array("A"))) == 'String'
    assert str(RuntimeType.infer(np.array(["A", "B"]))) == 'Vec<String>'


def test_typing_infer_to_string():
    # Currently these return actual strings, which we test with `is`.
    # The plan is to change the response type to be consistent:
    # https://github.com/opendp/opendp/issues/1665
    assert RuntimeType.infer(23) is 'i32' # noqa: F632
    assert RuntimeType.infer(12.) is 'f64' # noqa: F632
    assert RuntimeType.infer('hello') is 'String' # noqa: F632
    assert RuntimeType.infer(lambda: True) is 'CallbackFn' # noqa: F632    
    assert RuntimeType.infer(object(), py_object=True) is 'ExtrinsicObject' # noqa: F632
    

def test_typing_infer_to_object():
    # With py_object=True, it can fall back to a more general type:
    assert RuntimeType.infer([1, True], py_object=True) == 'Vec<ExtrinsicObject>'
    # Without py_object=True, it fails:
    with pytest.raises(TypeError, match=re.escape("elements must be homogeneously typed")):
        RuntimeType.infer([1, True])

    infer_vec = RuntimeType.infer(["A", "B"])
    assert infer_vec == 'Vec<String>'
    assert isinstance(infer_vec, RuntimeType)

    infer_tuple = RuntimeType.infer((12., True, "A"))
    assert infer_tuple == '(f64, bool, String)'
    assert isinstance(infer_tuple, RuntimeType)

    with pytest.raises(UnknownTypeException, match=re.escape("<class 'object'>")):
        RuntimeType.infer(object())
    
    with pytest.raises(UnknownTypeException, match=re.escape("Type of Option cannot be inferred from None")):
        RuntimeType.infer(None)
    with pytest.raises(UnknownTypeException, match=re.escape('Cannot infer atomic type when empty')):
        RuntimeType.infer([])

def test_typing_parse():
    assert str(RuntimeType.parse(Tuple[int, float])) == "(i32, f64)" # type: ignore[arg-type]
    assert str(RuntimeType.parse(tuple[int, float])) == "(i32, f64)" # type: ignore[arg-type]
    assert str(RuntimeType.parse(Tuple[int, Tuple[str]])) == "(i32, (String))" # type: ignore[arg-type]
    assert str(RuntimeType.parse(tuple[int, tuple[str]])) == "(i32, (String))" # type: ignore[arg-type]
    assert str(RuntimeType.parse(List[int])) == "Vec<i32>"
    assert str(RuntimeType.parse(list[int])) == "Vec<i32>"
    assert str(RuntimeType.parse(List[List[str]])) == "Vec<Vec<String>>"
    assert str(RuntimeType.parse(list[list[str]])) == "Vec<Vec<String>>"
    assert str(RuntimeType.parse((List[int], (int, bool)))) == '(Vec<i32>, (i32, bool))'
    assert str(RuntimeType.parse((list[int], (int, bool)))) == '(Vec<i32>, (i32, bool))'
    assert isinstance(RuntimeType.parse('L1Distance<f64>'), SensitivityMetric)
    with pytest.raises(UnknownTypeException):
        RuntimeType.parse(list[Any])


def test_sensitivity():
    assert isinstance(L1Distance[int], SensitivityMetric)
    assert not isinstance(RuntimeType.parse('(f32)'), SensitivityMetric)
    assert str(L1Distance[int]) == "L1Distance<i32>"
    assert L1Distance[int] != {}


def test_tuples():
    assert str(RuntimeType.parse((float, int))) == "(f64, i32)"
    assert str(RuntimeType.parse(('f32', (int, 'L1Distance<i32>')))) == '(f32, (i32, L1Distance<i32>))'
    assert str(RuntimeType.parse(('f32', (int, L2Distance[float])))) == '(f32, (i32, L2Distance<f64>))'


def test_c():
    """test that c_type strings are not mangled"""
    rt_type = RuntimeType.parse('FfiResult<void *>')
    assert isinstance(rt_type, RuntimeType)
    assert str(rt_type) == 'FfiResult<void *>'
    assert rt_type.args[0] == 'void *'


def test_feature_fails():
    with pytest.raises(OpenDPException, match=re.escape('enable_features("A", "Z")')):
        assert_features("A", "Z")


def test_set_feature():
    enable_features("A")
    assert_features("A")

    disable_features("A")
    assert "A" not in GLOBAL_FEATURES


def test_default_float_type():
    assert RuntimeType.parse(float) == f64

    set_default_float_type(f64)
    assert RuntimeType.parse(float) == f64

    # Can't set to f32 because debug binary has fewer types.

def test_runtime_type_hash():
    assert {Vec[int]} == {RuntimeType.parse("Vec<int>")}


disallowed_int_default_types = set([i128, u128, isize])

@pytest.mark.parametrize('integer_type', set(_INTEGER_TYPES) - disallowed_int_default_types)
def test_default_int_type(integer_type):
    assert RuntimeType.parse(int) == i32

    set_default_int_type(integer_type)
    assert RuntimeType.parse(int) == integer_type

    set_default_int_type(i32)


@pytest.mark.parametrize('integer_type', disallowed_int_default_types)
def test_disallowed_default_int_type(integer_type):
    assert RuntimeType.parse(int) == i32

    with pytest.raises(AssertionError):
        set_default_int_type(integer_type)

    set_default_int_type(i32)

# for a more thorough manual test of the set_default_int_type and set_default_float_type functions:
# 1. recompile with --release mode
# 2. set OPENDP_TEST_RELEASE
# 3. uncomment these commands
#    set_default_int_type(i64)
#    set_default_float_type(f32)
# 4. run pytest, and sanity check the failures



# type: ignore
def sample_bernoulli_exp(x) -> bool:
    while x >= 1:
        if sample_bernoulli_exp1(1):  # |\label{line:B_i}|
            x -= 1
        else: 
            return False
    return sample_bernoulli_exp1(x)   # |\label{line:C}|



# type: ignore
def make_row_by_row_fallible(
    input_domain: DI, 
    input_metric: M, 
    output_row_domain: DO, 
    # a function from input domain's row type to output domain's row type
    row_function: Callable([[DI_RowDomain_Carrier], DO_RowDomain_Carrier])
) -> Transformation:
    
    # where .translate is defined by the RowByRowDomain trait
    output_domain = input_domain.translate(output_row_domain)

    def function(data: DI_Carrier) -> DO_Carrier:
        # where .apply_rows is defined by the RowByRowDomain trait
        return DI.apply_rows(data, row_function)

    stability_map = new_stability_map_from_constant(1) # |\label{line:stability-map}|

    return Transformation(
        input_domain, output_domain, function,
        input_metric, input_metric, stability_map)


# type: ignore
def make_canonical_noise(
    input_domain: AtomDomain[f64],
    input_metric: AbsoluteDistance[f64],
    d_in: f64,
    d_out: tuple[f64, f64],
):
    assert not input_domain.nan(), "input data must be non-nan" # `\label{non-nan}`
    assert not d_in.is_sign_negative() and d_in.is_finite() # `\label{sensitivity-check}`

    tradeoff, fixed_point = approximate_to_tradeoff(d_out)
    r_d_in = RBig.try_from(d_in)

    def function(arg: f64) -> f64:  # `\label{function}`
        try:  # `\label{try-cast-arg}`
            arg = RBig.try_from(arg)
        except Exception:
            arg = RBig(0)
        
        canonical_rv = CanonicalRV(  # `\label{canonical-rv}`
            shift=arg, scale=r_d_in, tradeoff=tradeoff, fixed_point=fixed_point
        )
        return PartialSample.new(canonical_rv).value()  # `\label{sample-value}`

    def privacy_map(d_in_p: f64) -> f64:  # `\label{privacy-map}`
        assert 0 <= d_in_p <= d_in
        if d_in == 0:
            return (0.0, 0.0)
        return d_out

    return Measurement.new(
        input_domain,
        function,
        input_metric,
        output_measure=approximate(max_divergence()),
        privacy_map=privacy_map,
    )



# type: ignore
# analogous to impl MakeNoise<VectorDomain<AtomDomain<IBig>>, MI, MO> for RV in Rust
class RV:
    def make_noise_threshold(
        self,
        input_space: tuple[MapDomain[AtomDomain[TK], AtomDomain[IBig]], MI],
        threshold: IBig,
    ) -> Measurement[
        MapDomain[AtomDomain[TK], AtomDomain[IBig]], HashMap[TK, IBig], MI, MO
    ]:
        input_domain, input_metric = input_space
        output_measure = MO.default()
        threshold_magnitude = threshold.into_parts()[1] # |\label{line:threshold-mag}|
        privacy_map = self.noise_threshold_privacy_map( # |\label{line:privacy-map}|
            input_metric, output_measure, threshold_magnitude
        )

        match threshold.sign():
            case Sign.Positive:
                inner = Ordering.Less
            case Sign.Negative:
                inner = Ordering.Greater

        def function(data: HashMap[TK, IBig]) -> HashMap[TK, IBig]:
            out = []
            for k, v in data.items():
                v = self.sample(v) # |\label{line:sample}|

                if v.cmp(threshold) != inner:
                    out.append((k, v))
            # shuffle the output to avoid leaking the order of the input
            random.shuffle(out) # |\label{line:shuffle}|
            return dict(out)

        return Measurement.new(
            input_domain,
            Function.new_fallible(function),
            input_metric,
            output_measure,
            privacy_map,
        )



# type: ignore
def match_datetime_component(
    temporal_function: TemporalFunction,
) -> DataType | None:
    return ({
        Millennium: DataType.UInt32,
        Century: DataType.Int32,
        Year: DataType.Int32,
        IsoYear: DataType.Int32,
        Quarter: DataType.Int8,
        Month: DataType.Int8,
        Week: DataType.Int8,
        WeekDay: DataType.Int8,
        Day: DataType.Int8,
        OrdinalDay: DataType.Int16,
        Hour: DataType.Int8,
        Minute: DataType.Int8,
        Second: DataType.Int8,
        Millisecond: DataType.Int32,
        Microsecond: DataType.Int32,
        Nanosecond: DataType.Int32,
    }).get(temporal_function)




# type: ignore
def binary_search_by_mut(
    x: list[T],
    f: Callable[[T], Literal["less", "greater"]],
) -> int:
    size = x.len()
    if size == 0:  # |\label{fn-zero-size}|
        return 0
    
    base = 0 # |\label{fn-base}|

    while size > 1:
        half = size // 2
        mid = base + half

        cmp = f(x[mid])
        base = base if "greater" == cmp else mid

        size -= half

    cmp = f(x[base])
    return base + int(cmp == "less")



# Auto-generated. Do not edit!
'''
The ``internal`` module provides functions that can be used to construct library primitives without the use of the "honest-but-curious" flag.
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401
from opendp.core import *
from opendp.domains import *
from opendp.metrics import *
from opendp.measures import *
__all__ = [
    "_extrinsic_distance",
    "_extrinsic_divergence",
    "_extrinsic_domain",
    "_make_measurement",
    "_make_transformation",
    "_new_pure_function"
]


def _extrinsic_distance(
    identifier: str,
    descriptor = None
) -> Metric:
    r"""Construct a new ExtrinsicDistance.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `user_distance`.

    See `user_distance` for correct usage of this function.

    .. end-markdown

    :param identifier: A string description of the metric.
    :type identifier: str
    :param descriptor: Additional constraints on the domain.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=ctypes.c_char_p, type_name=None)
    c_descriptor = py_to_c(descriptor, c_type=ExtrinsicObjectPtr, type_name="ExtrinsicObject")

    # Call library function.
    lib_function = lib.opendp_internal___extrinsic_distance
    lib_function.argtypes = [ctypes.c_char_p, ExtrinsicObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier, c_descriptor), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': '_extrinsic_distance',
            '__module__': 'internal',
            '__kwargs__': {
                'identifier': identifier, 'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _extrinsic_divergence(
    descriptor: str
) -> Measure:
    r"""Construct a new ExtrinsicDivergence, a privacy measure defined from a bindings language.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `user_divergence`.

    See `user_divergence` for correct usage and proof definition for this function.

    .. end-markdown

    :param descriptor: A string description of the privacy measure.
    :type descriptor: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_descriptor = py_to_c(descriptor, c_type=ctypes.c_char_p, type_name="String")

    # Call library function.
    lib_function = lib.opendp_internal___extrinsic_divergence
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_descriptor), Measure))
    try:
        output.__opendp_dict__ = {
            '__function__': '_extrinsic_divergence',
            '__module__': 'internal',
            '__kwargs__': {
                'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _extrinsic_domain(
    identifier: str,
    member,
    descriptor = None
) -> Domain:
    r"""Construct a new ExtrinsicDomain.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `user_domain`.

    See `user_domain` for correct usage and proof definition for this function.

    .. end-markdown

    :param identifier: A string description of the data domain.
    :type identifier: str
    :param member: A function used to test if a value is a member of the data domain.
    :param descriptor: Additional constraints on the domain.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=ctypes.c_char_p, type_name=None)
    c_member = py_to_c(member, c_type=CallbackFnPtr, type_name="bool")
    c_descriptor = py_to_c(descriptor, c_type=ExtrinsicObjectPtr, type_name="ExtrinsicObject")

    # Call library function.
    lib_function = lib.opendp_internal___extrinsic_domain
    lib_function.argtypes = [ctypes.c_char_p, CallbackFnPtr, ExtrinsicObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier, c_member, c_descriptor), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': '_extrinsic_domain',
            '__module__': 'internal',
            '__kwargs__': {
                'identifier': identifier, 'member': member, 'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _make_measurement(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    function,
    privacy_map,
    TO: RuntimeTypeDescriptor = "ExtrinsicObject"
) -> Measurement:
    r"""Construct a Measurement from user-defined callbacks.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `make_user_measurement`.

    See `make_user_measurement` for correct usage and proof definition for this function.

    **Supporting Elements:**

    * Input Domain:   `AnyDomain`
    * Output Type:    `AnyMetric`
    * Input Metric:   `AnyMeasure`
    * Output Measure: `AnyObject`

    .. end-markdown

    :param input_domain: A domain describing the set of valid inputs for the function.
    :type input_domain: Domain
    :param input_metric: The metric from which distances between adjacent inputs are measured.
    :type input_metric: Metric
    :param output_measure: The measure from which distances between adjacent output distributions are measured.
    :type output_measure: Measure
    :param function: A function mapping data from ``input_domain`` to a release of type ``TO``.
    :param privacy_map: A function mapping distances from ``input_metric`` to ``output_measure``.
    :param TO: The data type of outputs from the function.
    :type TO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    TO = RuntimeType.parse(type_name=TO)

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name="AnyDomain")
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name="AnyMetric")
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name="AnyMeasure")
    c_function = py_to_c(function, c_type=CallbackFnPtr, type_name=pass_through(TO))
    c_privacy_map = py_to_c(privacy_map, c_type=CallbackFnPtr, type_name=measure_distance_type(output_measure))
    c_TO = py_to_c(TO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_internal___make_measurement
    lib_function.argtypes = [Domain, Metric, Measure, CallbackFnPtr, CallbackFnPtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_function, c_privacy_map, c_TO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': '_make_measurement',
            '__module__': 'internal',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'function': function, 'privacy_map': privacy_map, 'TO': TO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _make_transformation(
    input_domain: Domain,
    input_metric: Metric,
    output_domain: Domain,
    output_metric: Metric,
    function,
    stability_map
) -> Transformation:
    r"""Construct a Transformation from user-defined callbacks.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `make_user_transformation`.

    See `make_user_transformation` for correct usage and proof definition for this function.

    .. end-markdown

    :param input_domain: A domain describing the set of valid inputs for the function.
    :type input_domain: Domain
    :param input_metric: The metric from which distances between adjacent inputs are measured.
    :type input_metric: Metric
    :param output_domain: A domain describing the set of valid outputs of the function.
    :type output_domain: Domain
    :param output_metric: The metric from which distances between outputs of adjacent inputs are measured.
    :type output_metric: Metric
    :param function: A function mapping data from ``input_domain`` to ``output_domain``.
    :param stability_map: A function mapping distances from ``input_metric`` to ``output_metric``.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name="AnyDomain")
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name="AnyMetric")
    c_output_domain = py_to_c(output_domain, c_type=Domain, type_name="AnyDomain")
    c_output_metric = py_to_c(output_metric, c_type=Metric, type_name="AnyMetric")
    c_function = py_to_c(function, c_type=CallbackFnPtr, type_name=domain_carrier_type(output_domain))
    c_stability_map = py_to_c(stability_map, c_type=CallbackFnPtr, type_name=metric_distance_type(output_metric))

    # Call library function.
    lib_function = lib.opendp_internal___make_transformation
    lib_function.argtypes = [Domain, Metric, Domain, Metric, CallbackFnPtr, CallbackFnPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_domain, c_output_metric, c_function, c_stability_map), Transformation))
    try:
        output.__opendp_dict__ = {
            '__function__': '_make_transformation',
            '__module__': 'internal',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_domain': output_domain, 'output_metric': output_metric, 'function': function, 'stability_map': stability_map
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _new_pure_function(
    function,
    TO: RuntimeTypeDescriptor = "ExtrinsicObject"
) -> Function:
    r"""Construct a Function from a user-defined callback.
    This is meant for internal use, as it does not require "honest-but-curious",
    unlike `new_function`.

    See `new_function` for correct usage and proof definition for this function.


    Required features: `contrib`

    [_new_pure_function in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/internal/fn._new_pure_function.html)

    .. end-markdown

    :param function: A function mapping data to a value of type ``TO``
    :param TO: Output Type
    :type TO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    TO = RuntimeType.parse(type_name=TO)

    # Convert arguments to c types.
    c_function = py_to_c(function, c_type=CallbackFnPtr, type_name=pass_through(TO))
    c_TO = py_to_c(TO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_internal___new_pure_function
    lib_function.argtypes = [CallbackFnPtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_function, c_TO), Function))
    try:
        output.__opendp_dict__ = {
            '__function__': '_new_pure_function',
            '__module__': 'internal',
            '__kwargs__': {
                'function': function, 'TO': TO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output
